# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Pipeline helper class to create pipelines loading modules from a flexible manifest.
"""
from abc import abstractmethod, ABC
import logging
from packaging.specifiers import SpecifierSet
from packaging.version import Version, InvalidVersion

from azure.ml.component import Component
from azure.ml.component._core._component_definition import ComponentDefinition

from shrike.pipeline.v1.aml_connect import current_workspace
from shrike.pipeline.module_helper_base import AMLModuleLoaderBase

log = logging.getLogger(__name__)


class AMLModuleLoader(AMLModuleLoaderBase):
    def load_local_module_helper(self, module_spec_path):
        return Component.from_yaml(yaml_file=module_spec_path)

    def solve_module_version_and_load(
        self, module_name, module_version, module_cache_key, registry=None
    ):
        """Loads module class if exists

        Args:
            module_name (str): name of the module to load
            module_version (str): version of the module to load
            module_cache_key (str): cache key of the module after loading
            registy (str): registry name if loading from a registry
        """
        module_version = self.solve_module_version(
            module_name,
            module_version,
            workspace=current_workspace() if registry is None else None,
            registry=registry,
        )

        loaded_module_class = Component.load(
            current_workspace(),
            name=module_name,
            version=module_version,
            registry=registry,
        )

        self.put_in_cache(module_cache_key, loaded_module_class)
        return loaded_module_class

    def solve_module_version(
        self, module_name, module_version, workspace=None, registry=None
    ):
        if module_version is None:
            return module_version

        try:
            module_version_PEP440 = Version(module_version)
            if str(module_version_PEP440) != module_version:
                log.warning(
                    "We suggest adopting PEP440 versioning for your component {module_name}!"
                )

        except InvalidVersion as e:
            log.info(f"{module_version} is a version constraint. Try to solve it...")

            spec = SpecifierSet(module_version)
            if registry:
                components = ComponentDefinition.list(
                    registry_name=registry, name=module_name
                )
            elif workspace:
                components = ComponentDefinition.list(
                    workspace=workspace, name=module_name
                )

            versions = []
            for component in components:
                version = component.version
                version_PEP440 = Version(version)
                if str(version_PEP440) != version:
                    log.warning(
                        f"Version {version} does not follow PEP440 versioning, skipping ..."
                    )
                else:
                    versions.append(version_PEP440)

            compatible_versions = list(spec.filter(versions))
            if compatible_versions:
                module_version = max(compatible_versions)
                log.info(f"Solved version for {module_name} is {module_version}.")

            else:
                raise ValueError(
                    f"No version exists for the constraint {module_version}. Existing versions: {versions}"
                )

        return str(module_version)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
import os
import sys
import multiprocessing
import collections
import jsonpath_ng
import chardet
import re
from typing import List, Set, Union
import shutil
import subprocess
from ruamel.yaml import YAML
from git import Repo, InvalidGitRepositoryError, NoSuchPathError

from shrike.build.core.command_line import Command
from shrike.build.utils.utils import (
    create_catalog_stub,
    add_file_to_catalog,
    write_two_catalog_files,
    delete_two_catalog_files,
    get_component_name,
)
from pathlib import Path
import yaml
import urllib.parse
import uuid
from urllib.parse import urlparse
from enum import Enum

log = logging.getLogger(__name__)

ALLOWED_CONTAINER_REGISTRIES = ["polymerprod.azurecr.io"]
ALLOWED_PACKAGE_FEEDS = [
    "https://o365exchange.pkgs.visualstudio.com/_packaging/PolymerPythonPackages/pypi/simple/",
    "https://o365exchange.pkgs.visualstudio.com/_packaging/O365PythonPackages/pypi/simple/"
]


class RuntimeEnvironment(Enum):
    AZURE_DEVOPS_BUILD = "Azure DevOps Build"
    GITHUB_ACTION = "GitHub Action"
    OTHER = "Other"


class OperatingSystem(Enum):
    WINDOWS = "Windows"
    LINUX = "Linux"
    OTHER = "Other"


class TargetType(Enum):
    ADDITIONAL_INCLUDES = "additional_includes"
    DEPENDENCY_HINTS = "dependency_hints"


class ActionType(Enum):
    VALIDATE = "validate"
    BUILD = "build"


class Prepare(Command):
    def __init__(self):
        super().__init__()
        self._component_statuses = {}

    def folder_path(self, file: str) -> str:
        """
        Return the normalized path of the directory containing a file.
        """
        return self.normalize_path(Path(file).parent, directory=True)

    def all_files_in_snapshot(self, manifest: str) -> List[str]:
        """
        Return a list of all normalized files in the snapshot. The input
        (`manifest`) is assumed to be some AML-style component
        spec, in the "root" of the
        snapshot.
        """
        folder_path = self.folder_path(manifest)
        log.info("Absolute path for current component is: " + folder_path)

        # Generate a list of all files in this components folder (including subdirectories)
        rv = []

        # Make sure we pick up Linux-style "hidden" files like .amlignore and
        # hidden "directories", as well as hidden files in hidden directories.
        # https://stackoverflow.com/a/65205404
        # https://stackoverflow.com/a/41447012
        for root, _, file_paths in os.walk(folder_path):
            for file in file_paths:
                file_path = os.path.join(root, file)
                normalized_path = self.normalize_path(file_path)
                rv.append(normalized_path)

        return rv

    def process_all_components(self, files: List[str], action_type) -> List[str]:  # type: ignore
        """
        Depending on the 'action_type' ('build' or 'validate'), run
        build_each_components() or validate_each_component()
        in parallel with nb_cores threads.
        Returns the list of "built" component files, or an empty list
        in the validate case.
        """
        rv = []
        if files:
            nb_cores = self.nb_cores

            if nb_cores == 1:
                for component in files:
                    if action_type == ActionType.BUILD:
                        build_rv = self.build_each_components(component)
                        if build_rv:
                            # if build fails, build_rv is None
                            rv.append(build_rv)
                    elif action_type == ActionType.VALIDATE:
                        self.validate_each_components(component)
                    else:
                        log.error(
                            f"Unknown value for action_type: '{action_type}'. It should be either '{ActionType.BUILD}' or '{ActionType.VALIDATE}'"
                        )
            else:
                log.info(
                    f"Batch component processing with {nb_cores} threads. Action: '{action_type}'"
                )
                pool = multiprocessing.Pool(processes=nb_cores)
                if action_type == ActionType.BUILD:
                    rv = pool.map(self.build_each_components, files)
                elif action_type == ActionType.VALIDATE:
                    pool.map(self.validate_each_components, files)
                pool.close()
                pool.join()
            return rv  # type: ignore

    def build_each_components(self, component):
        """
        For one of component specification file, run `mldesigner compile`,
        and register the status (+ register error if build failed).
        """
        path = Path(component)
        res = subprocess.run(
            args=f"mldesigner compile --source {component} --output {path.parent}/.build".split(),
            cwd = self.config.working_directory,
            stdout = subprocess.PIPE,
            stderr = subprocess.PIPE
        )
        stderr = str(res.stderr, encoding="utf-8", errors="ignore")
        success_log = "Mldesigner has compiled 1 components from 1 files."
        build_component_success = res.returncode == 0 and success_log in stderr

        if build_component_success:
            log.info(f"Component {component} is built.")
            with open(component, "r") as spec_file:
                spec = YAML(typ="safe").load(spec_file)
            component_name = get_component_name(spec, path)
            log.info(f"component_name {component_name}")
            return str(path.parent / ".build" / component_name /path.name)
        else:
            # if build fails, the path above won't exist, so no return
            self.register_error(f"Error when building component {component}. Logs: {stderr}")


    def create_catalog_files(self, files: List[str]):
        """
        Create the appropriate kind of catalog file(s), using the configured
        method ("aml" or "aether").
        """
        signing_mode = self.config.signing_mode

        if signing_mode == "aml":
            self.create_catalog_files_for_aml(files)
        else:
            raise ValueError(f"Invalid signing_mode provided: '{signing_mode}'")

    def create_catalog_files_for_aml(self, files: List[str]) -> None:
        """
        Create AML-friendly catalog.json and catalog.json.sig files, using
        SHA-256 hash.
        """

        # For each component spec file in the input list, we'll do the following...
        for f in files:
            log.info(f"Processing file {f}")
            component_folder_path = self.folder_path(f)

            # remove catalog files if already present
            log.info("Deleting old catalog files if present")
            delete_two_catalog_files(component_folder_path)

            files_for_catalog = self.all_files_in_snapshot(f)
            log.info("The following list of files will be added to the catalog.")
            log.info(files_for_catalog)

            # Prepare the catlog stub: {'HashAlgorithm': 'SHA256', 'CatalogItems': {}}
            catalog = create_catalog_stub()

            # Add an entry to the catalog for each file
            for file_for_catalog in files_for_catalog:
                catalog = add_file_to_catalog(
                    file_for_catalog, catalog, component_folder_path
                )

            # order the CatalogItems dictionary
            catalog["CatalogItems"] = collections.OrderedDict(
                sorted(catalog["CatalogItems"].items())
            )

            # Write the 2 catalog files
            log.info(catalog)
            write_two_catalog_files(catalog, component_folder_path)
            log.info("Finished creating catalog files.")

    def find_component_specification_files(self) -> List[str]:
        """
        Find the list of "active" component specification files using the
        configured method ("all" or "smart").
        """
        activation_method = self.config.activation_method

        if activation_method == "all":
            rv = self.find_component_specification_files_using_all()
        elif activation_method == "smart":
            rv = self.find_component_specification_files_using_smart()
        else:
            raise ValueError(
                f"Invalid activation_method provided: '{activation_method}'"
            )

        return rv

    def add_repo_and_last_pr_to_tags(self, files: List[str]) -> List[str]:
        [repo, current_branch, compliant_branch] = self.identify_repo_and_branches()
        repo_path = repo.remotes.origin.url

        for file in files:
            with open(file, "r") as spec_file:
                spec = yaml.load(spec_file, Loader=yaml.FullLoader)
            if not isinstance(spec, dict):
                continue
            last_commit = next(repo.iter_commits(paths=file, max_count=1))
            last_commit_id = last_commit.hexsha

            last_commit_message = last_commit.summary
            path_to_component = os.path.relpath(
                os.path.split(file)[0], repo.working_dir
            )
            link_to_commit = (
                repo_path
                + "?version=GC"
                + last_commit_id
                + "&path="
                + urllib.parse.quote(path_to_component, safe="")
            )
            # update description
            cur_description = spec.get("description")
            new_description = f"[link to commit]({link_to_commit})"
            if cur_description:
                spec["description"] = cur_description + "\n----\n" + new_description
            else:
                spec["description"] = new_description
            # update tags
            new_tag = {
                "repo": repo_path,
                "last_commit_id": last_commit_id,
                "last_commit_message": last_commit_message,
                "path_to_component": path_to_component,
            }
            cur_tag = spec.get("tags")
            if cur_tag is None:
                cur_tag = new_tag
            else:
                cur_tag.update(new_tag)
            spec["tags"] = cur_tag
            with open(file, "w") as spec_file:
                yaml.dump(spec, spec_file, sort_keys=False)
        return files

    def find_component_specification_files_using_all(self, dir=None) -> List[str]:
        """
        Find all component specification files in the configured working
        directory matching the configured glob. Return the absolute paths
        of these files in the format of a list of string.
        """
        if dir is None:
            dir = self.config.working_directory
        all_spec_yaml_files_absolute_paths = [
            str(p.absolute())
            for p in Path(dir).glob(self.config.component_specification_glob)
            if '.build' not in p.parts
        ]

        return all_spec_yaml_files_absolute_paths

    def find_component_specification_files_using_smart(self) -> List[str]:
        """
        This function returns the list of components (as a list of absolute paths) potentially affected by the latest commit.
        """
        log.info(
            "Determining which components are potentially affected by the current change."
        )
        [repo, current_branch, compliant_branch] = self.identify_repo_and_branches()
        modified_files = self.get_modified_files(repo, current_branch, compliant_branch)
        active_components = self.infer_active_components_from_modified_files(
            modified_files
        )
        return active_components

    def identify_repo_and_branches(self):
        """
        This function returns the current repository, along with the name of the current and compliant branches [repo, current_branch, compliant_branch]. Throws if no repo can be found.
        """
        # identify the repository
        curr_path = Path(self.config.working_directory).resolve()
        try:
            repo = Repo(curr_path, search_parent_directories=True)
            log.info("Found a valid repository in " + str(repo.git_dir))
        except (InvalidGitRepositoryError, NoSuchPathError):
            message = (
                str(curr_path)
                + " or its parents do not contain a valid repo path or cannot be accessed."
            )
            raise Exception(message)
        try:
            current_branch = str(
                repo.head.ref
            )  # when running from our build the repo head is detached so this will throw an exception
        except TypeError:
            current_branch = os.environ.get("BUILD_SOURCEBRANCH") or os.environ.get(
                "GITHUB_REF"
            )
        log.info("The current branch is: '" + str(current_branch) + "'.")
        # Identify the compliant branch
        if not (self.config.compliant_branch.startswith("^refs/heads/")) or not (
            self.config.compliant_branch.endswith("$")
        ):
            raise Exception(
                "The name of the compliant branch found in the config file should start with '^refs/heads/' and end with '$'. Currently it is: '"
                + self.config.compliant_branch
                + "'."
            )
        else:
            compliant_branch = self.config.compliant_branch.replace("^refs/heads/", "")[
                0:-1
            ]
        log.info("The compliant branch is: '" + compliant_branch + "'.")
        return [repo, current_branch, compliant_branch]

    def get_modified_files(self, repo, current_branch, compliant_branch) -> Set[str]:
        """
        This function returns the paths of files that have been modified. 3 scenarios are supported.\n
        1/ 'Build - before Merge'; when the 'prepare' command is run as part of a build, but before the actual merge (in this case, the name of the current branch starts with 'refs/pull/' - this is the default Azure DevOps behavior).\n
        2/ 'Build - after Merge'; when the 'prepare' command is run as part of a build, after the actual merge (in this case, the name of the current branch is the same as the name of the compliant branch).\n
        3/ 'Manual'; when the prepare command is run manually (typically before publishing the PR).
        """
        # identify the 2 relevant commits based on the use case
        current_commit, previous_commit = self.get_relevant_commits(
            repo, current_branch, compliant_branch
        )

        # take the actual diff
        diff = self.get_diff_between_commits(current_commit, previous_commit)

        # process the diff object to obtain a list of paths
        res = self.extract_paths_from_diff(
            diff,
            repo_working_dir=repo.working_dir,
            repo_working_tree_dir=repo.working_tree_dir,
            repo_git_dir=repo.git_dir,
        )
        return res

    def get_relevant_commits(self, repo, current_branch, compliant_branch):
        """
        This function returns the commits required to compute the list of files that have been modified. 3 scenarios are supported.\n
        1/ 'Build - before Merge'; when the 'prepare' command is run as part of a build, but before the actual merge (in this case, the name of the current branch starts with 'refs/pull/' - this is the default Azure DevOps behavior).\n
        2/ 'Build - after Merge'; when the 'prepare' command is run as part of a build, after the actual merge (in this case, the name of the current branch is the same as the name of the compliant branch).\n
        3/ 'Manual'; when the prepare command is run manually (typically before publishing the PR).
        """
        # Grab the diff differently depending on the scenario
        if current_branch.replace("refs/heads/", "") == compliant_branch:
            # 'Build - after Merge' case: we will take the diff between the
            # tree of the latest commit to the compliant branch, and the tree
            # of the previous commit to the compliant branch corresponding to a
            # PR (we assume the commit summary starts with 'Merged PR')
            log.info(
                "We are in the 'Build - after Merge' case (the current branch is the compliant branch)."
            )
            current_commit = self.get_compliant_commit_corresponding_to_pull_request(
                repo, compliant_branch
            )
            self.log_commit_info(current_commit, "Current commit to compliant branch")
            previous_commit = (
                self.get_previous_compliant_commit_corresponding_to_pull_request(
                    current_commit,
                    consider_current_commit=False,
                )
            )
            self.log_commit_info(
                previous_commit, "Previous PR commit to compliant branch"
            )
        elif current_branch.startswith("refs/pull/"):
            # 'Build - before Merge': we will take the diff between the tree of
            # the current commit, and the tree of the previous commit to the
            # compliant branch corresponding to a PR (we assume the commit
            # summary starts with 'Merged PR')
            log.info(
                "We are in the 'Build - before Merge' case (the current branch is not the compliant branch and its name starts with 'refs/pull/')."
            )
            current_commit = repo.commit()
            self.log_commit_info(current_commit, "Current commit to current branch")
            latest_commit_to_compliant_branch = repo.remotes.origin.refs[
                compliant_branch
            ].commit
            previous_commit = (
                self.get_previous_compliant_commit_corresponding_to_pull_request(
                    latest_commit_to_compliant_branch,
                    consider_current_commit=True,
                )
            )
            self.log_commit_info(
                previous_commit, "Previous PR commit to compliant branch"
            )
        else:
            # 'Manual' Case: we will take the diff between the current branch
            # and the compliant branch (we're assuming the compliant branch is
            # locally up to date here)
            log.info(
                "We are in the 'Manual' case (the current branch is NOT the compliant branch and its name does not start with 'refs/pull/')."
            )
            try:
                current_commit = repo.heads[
                    current_branch
                ].commit  # this won't work when running the Manual case from the DevOps portal, but the below will
            except (IndexError, AttributeError):
                current_commit = repo.commit()
            self.log_commit_info(current_commit, "Current commit to current branch")
            try:
                previous_commit = repo.heads[
                    compliant_branch
                ].commit  # this won't work when running the Manual case from the DevOps portal, but the below will
            except (IndexError, AttributeError):
                latest_commit_to_compliant_branch = repo.remotes.origin.refs[
                    compliant_branch
                ].commit
                previous_commit = (
                    self.get_previous_compliant_commit_corresponding_to_pull_request(
                        latest_commit_to_compliant_branch,
                        consider_current_commit=True,
                    )
                )
            self.log_commit_info(previous_commit, "Previous commit to compliant branch")

        return current_commit, previous_commit

    def get_diff_between_commits(self, current_commit, previous_commit):
        """Function that gets the diff between 2 commits."""
        # just use the 'diff' function from gitpython
        return current_commit.tree.diff(previous_commit.tree)

    def extract_paths_from_diff(
        self, diff, repo_working_dir, repo_working_tree_dir, repo_git_dir
    ):
        """Function that extracts the paths of the modified files from the diff between 2 commits."""
        res = set()
        # let's build a set with the paths of modified files found in the diff object
        log.debug("Working directory: " + self.config.working_directory)
        log.debug("repo.working_dir: " + repo_working_dir)
        log.debug("repo.working_tree_dir: " + repo_working_tree_dir)
        log.debug("repo.git_dir: " + repo_git_dir)
        for d in diff:
            log.debug("d.a_path: " + d.a_path)
            log.debug("Path(d.a_path).absolute(): " + str(Path(d.a_path).absolute()))
            log.debug("Path(d.a_path).resolve(): " + str(Path(d.a_path).resolve()))
            r_a = str(Path(repo_git_dir).parent / Path(d.a_path))
            res.add(r_a)
            r_b = str(Path(repo_git_dir).parent / Path(d.b_path))
            res.add(r_b)
        log.info("The list of modified files is:")
        log.info(res)
        return res

    def log_commit_info(self, commit, title) -> None:
        log.info(title + ":")
        log.info("Summary: " + commit.summary)
        log.info("Author: " + str(commit.author))
        log.info("Authored Date: " + str(commit.authored_date))

    def get_previous_compliant_commit_corresponding_to_pull_request(
        self, latest_commit, consider_current_commit
    ):
        """
        This function will return the previous commit in the repo corresponding to a PR (i.e. that starts with "Merged PR").
        If `consider_current_commit` is set to True, the `latest_commit` will be considered. If set to false, only previous commits will be considered.
        """
        target_string = "Merged PR"
        if consider_current_commit and latest_commit.summary.startswith(target_string):
            return latest_commit
        previous_commit = latest_commit
        for c in previous_commit.iter_parents():
            if c.summary.startswith(target_string):
                previous_commit = c
                break
        return previous_commit

    def get_compliant_commit_corresponding_to_pull_request(
        self, repo, compliant_branch
    ):
        """
        This function will return the most recent commit in the repo that truly corresponds to the triggered build. It is identified thanks to the 'Build.SourceVersionMessage' DevOps environment variable (see https://docs.microsoft.com/en-us/azure/devops/pipelines/build/variables?view=azure-devops&tabs=yaml) that contains the true commit message. This is used to address the race condition occurring when a commit sneaks in before the "prepare" step was run on the previous commit.
        """
        # this is the true commit message corresponding to the PR that triggered the build
        true_commit_message = self.get_true_commit_message()
        # this is the most recent commit
        current_commit = repo.remotes.origin.refs[compliant_branch].commit
        # if the most recent commit corresponds to the true commit message, then return it
        if true_commit_message.startswith(current_commit.summary):
            return current_commit
        # otherwise, let's iterate through the parents until we find it
        candidate_commit = current_commit
        for c in candidate_commit.iter_parents():
            if true_commit_message.startswith(c.summary):
                return c
        # if the corresponding commit cannot be found, return the most recent one and log a warning
        log.warning(
            "Could not find in the git repo the commit that triggered this PR. Returning the most recent but beware, the 'smart' mode likely will not work properly."
        )
        return current_commit

    def get_true_commit_message(self):
        return str(os.environ.get("BUILD_SOURCEVERSIONMESSAGE") or "NA")

    def infer_active_components_from_modified_files(self, modified_files) -> List[str]:
        """
        This function returns the list of components (as a list of directories paths) potentially affected by changes in the `modified_files`.
        """
        rv = []
        # We will go over components one by one
        all_components_in_repo = self.find_component_specification_files_using_all()
        log.info("List of all components in repo:")
        log.info(all_components_in_repo)
        for component in all_components_in_repo:
            if self.component_is_active(component, modified_files):
                rv.append(component)
        # No need to dedup rv since we are only considering components once
        log.info("The active components are:")
        log.info(rv)
        return rv

    def component_is_active(self, component, modified_files) -> bool:
        """
        This function returns True if any of the 'modified_files' potentially affects the 'component' (i.e. if it is directly in one of the 'component' subfolders, or if it is covered by the additional_includes files). If the component has been deleted, returns False.
        """
        log.info("Assessing whether component '" + component + "' is active...")
        # Let's first take care of the case where the component has been deleted
        if not (Path(component).exists()):
            return False
        # Let's grab the contents of the additional_includes file if it exists.
        component_additional_includes_contents = self.get_target_file_contents(
            component,
            TargetType.ADDITIONAL_INCLUDES,
        )
        # Let's grab the contents of the additional_includes file if it exists.
        component_dependency_hints_contents = self.get_target_file_contents(
            component,
            TargetType.DEPENDENCY_HINTS,
        )
        # loop over all modified files; if current file is in subfolder of component or covered by
        # additional includes or dependency hints, return True
        for modified_file in modified_files:
            if (
                self.is_in_subfolder(modified_file, component)
                or self.is_in_target_list(
                    modified_file,
                    TargetType.ADDITIONAL_INCLUDES,
                    component_additional_includes_contents,
                )
                or self.is_in_target_list(
                    modified_file,
                    TargetType.DEPENDENCY_HINTS,
                    component_dependency_hints_contents,
                )
            ):
                return True
        return False

    def get_target_file_contents(
        self, component, target_type
    ) -> Union[List[str], None]:
        component_target_file_contents = None
        # for depependency hints, we look globally first
        if target_type == TargetType.DEPENDENCY_HINTS:
            component_target_file_contents = self.get_global_dependency_hints_contents(
                component
            )
        # First we figure out the expected path of the additional_includes file
        component_target_file_path = self.get_theoretical_target_file_path(
            component, target_type
        )
        # And we load it if it exists.
        if Path(component_target_file_path).exists():
            rbfile = open(component_target_file_path, "rb").read()
            if chardet.detect(rbfile).get("encoding").lower() not in ["utf-8", "ascii"]:
                raise ValueError(
                    f"Encoding of a file: '{{spec_file_name}}.{target_type.value}' not supported, use UTF-8."
                )

            with open(component_target_file_path, "r") as component_target_file:
                if (target_type == TargetType.DEPENDENCY_HINTS) and (
                    component_target_file_contents is not None
                ):
                    component_target_file_contents += component_target_file.readlines()
                else:
                    component_target_file_contents = component_target_file.readlines()
        else:
            # If additional_includes doesn't exist we log a message explaining the expected name format
            if target_type == TargetType.ADDITIONAL_INCLUDES:
                log.info(
                    "No additional_includes file could be found for the component '"
                    + component
                    + "'. If you tried to create such a file, remember it should live next to the component spec file and should be named '{spec_file_name}.additional_includes'. "
                    + "For example, if the component spec file is named 'component_spec.yaml', the additional_includes file should be named 'component_spec.additional_includes'. In this specific case, the expected additional_includes file name is: '"
                    + component_target_file_path
                    + "'."
                )
                # Then we check whether there is an improperly named additional_includes file in the component folder, and if so we throw
                if self.check_for_wrongly_named_additional_includes(component):
                    raise ValueError(
                        "An additional_includes file which does not respect the naming pattern was found. Please rename this file. Remember it should live next to the component spec file and should be named '{spec_file_name}.additional_includes'."
                        + "For example, if the component spec file is named 'component_spec.yaml', the additional_includes file should be named 'component_spec.additional_includes'."
                    )
        # Before returning, we make the paths in the additional_includes file absolute
        if component_target_file_contents:
            for line_number in range(0, len(component_target_file_contents)):
                component_target_file_contents[line_number] = str(
                    Path(
                        os.path.join(
                            Path(component).parent,
                            component_target_file_contents[line_number].rstrip("\n"),
                        )
                    ).resolve()
                )
        return component_target_file_contents

    def get_global_dependency_hints_contents(self, component) -> Union[List[str], None]:
        if len(self.config.dependency_hints) > 0:
            global_dependency_hints_contents = []
            for (
                component_folder_paths,
                dependency_hints_paths,
            ) in self.config.dependency_hints.items():
                component_folder_absolute_paths = [
                    str(p.absolute().resolve())
                    for p in Path(self.config.working_directory).glob(
                        component_folder_paths
                    )
                ]
                if (
                    str(Path(component).parent.resolve())
                    in component_folder_absolute_paths
                ):
                    if not isinstance(dependency_hints_paths, list):
                        dependency_hints_paths = [dependency_hints_paths]
                    for dependency_hints_path in list(dependency_hints_paths):
                        global_dependency_hints_contents += [
                            str(p.absolute().resolve())
                            for p in Path(self.config.working_directory).glob(
                                dependency_hints_path
                            )
                        ]
            if len(global_dependency_hints_contents) > 0:
                return global_dependency_hints_contents
            else:
                return None
        else:
            return None

    def get_theoretical_target_file_path(self, component, target_type) -> str:
        """
        Returns the expected path of the 'target_type' file
        associated with the 'component'.
        """
        # First, we figure out the name of the target file, based on the component name
        component_name_without_extension = Path(component).name.split(".yaml")[0]
        # Then, we construct the expected path of the target file (see
        # https://componentsdk.azurewebsites.net/components/component-spec-topics/additional-includes.html
        # for the 'additional_includes' case)
        component_target_file_path = os.path.join(
            Path(component).parent,
            component_name_without_extension + "." + str(target_type.value),
        )
        return component_target_file_path

    def check_for_wrongly_named_additional_includes(self, component) -> bool:
        """
        Returns True if the component folder contains an improperly named additional_includes file
        i.e. a lonely additional_includes file without a corresponding spec.yaml
        """
        # grab all potential additional_includes files in the component folder
        potentially_wrongly_named_files = Path(component).parent.glob(
            "*.additional_includes*"
        )
        for potentially_wrongly_named_file in potentially_wrongly_named_files:
            # determine the expected location of the spec
            theoretical_component_path = (
                str(potentially_wrongly_named_file)[:-20] + ".yaml"
            )
            # check if spec exists
            if os.path.isfile(theoretical_component_path):
                continue
            else:
                # if not, we have a problem
                self.register_error(
                    f"Component folder {component} contains a lonely additional includes file at {potentially_wrongly_named_file}, missing component spec {theoretical_component_path}"
                )
                return True
        return False

    def is_in_subfolder(self, modified_file, component) -> bool:
        """
        This function returns True if 'modified_file' is in a subfolder of 'component' ('component' can be either the path to a file, or a directory). If the component has been deleted, returns False.
        """
        # Let's first take care of the case where the component has been deleted
        if not (Path(component).exists()):
            log.debug("'" + component + "' does not exist, returning False.")
            return False
        # Case where the component has not been deleted
        for parent in Path(modified_file).parents:
            if parent.exists():
                if Path(component).is_dir():
                    if parent.samefile(Path(component)):
                        log.info(
                            "'"
                            + modified_file
                            + " is in a subfolder of '"
                            + component
                            + "'."
                        )
                        return True
                else:
                    if parent.samefile(Path(component).parent):
                        log.info(
                            "'"
                            + modified_file
                            + " is in a subfolder of '"
                            + component
                            + "'."
                        )
                        return True
        log.debug(
            "'" + modified_file + " is NOT in a subfolder of '" + component + "'."
        )
        return False

    def is_in_target_list(
        self, modified_file, target_type, target_list_contents
    ) -> bool:
        """
        This function returns True if 'modified_file' is covered by the file
        'target_list_contents'. The 'target_type' can be either
        additional_includes or dependency_hints
        """
        # first tackle the trivial case of no target file
        if target_list_contents is None:
            log.debug(
                f"The component's target file ({target_type}) is empty, returning False."
            )
            return False
        # now the regular scenario
        for line in target_list_contents:
            # when the line from the target list is a file, we directly check its path against that of modified_file
            if Path(line).is_file():
                if str(Path(modified_file).resolve()) == str(
                    Path(line).resolve()
                ):  # can't use 'samefile' here because modified_file is not guaranteed to exist, we resolve the path and do basic == test
                    log.info(
                        "'"
                        + modified_file
                        + f" is directly listed in the {target_type} file."
                    )
                    return True
            # slightly more complicated case: when the line
            # in the target list is a directory, we can just
            # call the is_in_subfolder function
            # but first, we take care of the zipped folders
            if target_type == TargetType.ADDITIONAL_INCLUDES:
                if self.config.detect_changes_in_unzipped_folder:
                    split_line = os.path.splitext(line)
                    if split_line[1] == ".zip":
                        line = split_line[0]
            if Path(line).is_dir():
                if self.is_in_subfolder(modified_file, line):
                    log.info(
                        "'"
                        + modified_file
                        + f" is in one of the directories listed in the {target_type} file."
                    )
                    return True
        log.debug(
            "'"
            + modified_file
            + f" is NOT referenced by the {target_type} file (neither directly nor indirectly)."
        )
        return False

    def _create_dependencies_files(self, component_files) -> str:
        id = str(uuid.uuid4())
        path_to_dependencies_files = os.path.join(
            self.config.working_directory, "component_dependencies_" + id
        )
        log.info(
            f"Writing Python package dependencies to path {path_to_dependencies_files}"
        )
        os.makedirs(path_to_dependencies_files)
        for component in component_files:
            self._create_dependencies_files_for_single_component(
                component, path_to_dependencies_files
            )
        return id

    def _create_dependencies_files_for_single_component(
        self, component, path_to_dependencies_files
    ) -> None:
        component_repo = Path(component).parent
        with open(component, "r") as spec_file:
            spec = YAML(typ="safe").load(spec_file)
        component_name = get_component_name(spec, component)
        (
            pip_dependencies,
            conda_dependencies,
            _,
        ) = self._extract_dependencies_and_channels(component, component_name)

        if pip_dependencies or conda_dependencies:
            cur_path = os.path.join(path_to_dependencies_files, component_name)
            try:
                os.makedirs(cur_path)
            except FileExistsError:
                suffix = (
                    component_name
                    + "_"
                    + os.path.splitext(os.path.basename(component))[0]
                )
                cur_path = os.path.join(path_to_dependencies_files, suffix)
                os.makedirs(cur_path)
            if pip_dependencies:
                log.info(
                    f"Found pip dependencies for component {component_name} in {component_repo}. Writing to requirements.txt."
                )
                with open(os.path.join(cur_path, "requirements.txt"), "w") as file:
                    for req in pip_dependencies:
                        file.write(req)
                        if not req.endswith("\n"):
                            file.write("\n")
            if conda_dependencies:
                log.info(
                    f"Found conda dependencies for component {component_name} in {component_repo}. Writing to environment.yml."
                )
                with open(os.path.join(cur_path, "environment.yml"), "w") as file:
                    yaml.dump(conda_dependencies, file)

    def _extract_dependencies_and_channels(self, component, component_name) -> List[list]:
        component_repo = Path(component).parent
        build_folder = os.path.join(component_repo, ".build", component_name)
        if os.path.exists(build_folder):
            component_repo = build_folder
        with open(component, "r") as spec_file:
            spec = YAML(typ="safe").load(spec_file)
        pip_dependencies = []
        conda_dependencies = []
        conda_channels = []
        if "type" in spec and spec["type"] == "spark":
            if "conda_dependencies" in spec:
                conda_dependencies = spec["conda_dependencies"]
                pip_dependencies += self._extract_python_package_dependencies(
                    conda_dependencies
                )
                if "channels" in conda_dependencies:
                    conda_channels += conda_dependencies["channels"]
            if "environment" in spec and "conda_file" in spec["environment"]:
                conda_dependencies_file = spec["environment"]["conda_file"]
                try:
                    with open(
                        os.path.join(
                            component_repo, conda_dependencies_file
                        )
                    ) as file:
                        conda_dependencies = YAML(typ="safe").load(file)
                    pip_dependencies += self._extract_python_package_dependencies(
                        conda_dependencies
                    )
                    if "channels" in conda_dependencies:
                        conda_channels += conda_dependencies["channels"]
                except FileNotFoundError:
                    self.register_error(
                        f"The required conda_dependencies_file {conda_dependencies_file} does not exist in {component_repo}."
                    )
        if "environment" in spec:
            spec_environment = spec.get("environment")
            if "conda" in spec_environment:
                spec_conda = spec_environment["conda"]
                if "conda_dependencies" in spec_conda:
                    conda_dependencies = spec_conda["conda_dependencies"]
                    pip_dependencies += self._extract_python_package_dependencies(
                        conda_dependencies
                    )
                    if "channels" in conda_dependencies:
                        conda_channels += conda_dependencies["channels"]
                if "conda_dependencies_file" in spec_conda:
                    conda_dependencies_file = spec_conda["conda_dependencies_file"]
                    try:
                        with open(
                            os.path.join(
                                component_repo, conda_dependencies_file
                            )
                        ) as file:
                            conda_dependencies = YAML(typ="safe").load(file)
                        pip_dependencies += self._extract_python_package_dependencies(
                            conda_dependencies
                        )
                        if "channels" in conda_dependencies:
                            conda_channels += conda_dependencies["channels"]
                    except FileNotFoundError:
                        self.register_error(
                            f"The required conda_dependencies_file {conda_dependencies_file} does not exist in {component_repo}."
                        )
                if "pip_requirements_file" in spec_conda:
                    pip_requirements_file = spec_conda["pip_requirements_file"]
                    try:
                        with open(
                            os.path.join(
                                component_repo, spec_conda["pip_requirements_file"]
                            )
                        ) as file:
                            pip_dependencies += file.readlines()
                    except FileNotFoundError:
                        self.register_error(
                            f"The required pip_requirements_file {pip_requirements_file} does not exist in {component_repo}."
                        )
        return [pip_dependencies, conda_dependencies, conda_channels]

    def _extract_python_package_dependencies(self, conda_dependencies) -> List[str]:
        pip_dependencies = []
        if "dependencies" in conda_dependencies:
            dependencies = conda_dependencies.get("dependencies")
            for dependencies_item in dependencies:
                if isinstance(dependencies_item, dict) and "pip" in dependencies_item:
                    pip_dependencies = dependencies_item["pip"]
        return pip_dependencies

    def run_with_config(self):
        log.info("Running component preparation logic.")

        self.telemetry_logging(command="prepare")

        component_files = self.find_component_specification_files()
        if not self.config.suppress_adding_repo_pr_tags:
            try:
                component_files = self.add_repo_and_last_pr_to_tags(component_files)
            except StopIteration:
                log.warning(
                    "`add_repo_and_last_pr_to_tags` not successful. Please make sure your component files are in Git. Otherwise, please set `suppress_adding_repo_pr_tags` to True."
                )

        if self.config.signing_mode == "aml":
            self.ensure_component_cli_installed()
            self.attach_workspace()
            self.process_all_components(component_files, ActionType.VALIDATE)
            built_component_files = self.process_all_components(
                component_files, ActionType.BUILD
            )
        else:
            built_component_files = component_files

        # only call self.create_catalog_files if built_component_files not None
        log.info(f"Built component files: {built_component_files}")
        if built_component_files:
            self.create_catalog_files(built_component_files)

        self._create_dependencies_files(component_files)

        self.add_conda_to_system_path()

    def validate_each_components(self, component) -> None:
        """
        For one of component specification file, run `az ml component validate`,
        run compliance and customized validation if enabled,
        and register the status (+ register error if validation failed).
        """
        validate_command = f"ml component validate --file {component}"
        if self.config.verbose:
            validate_command +=  " --debug"
        validate_component_success = self.execute_azure_cli_command(
            validate_command, stderr_is_failure=False, check_stdout_result=True, ignore_blowfish=True
        )
        compliance_validation_success = True
        customized_validation_success = True
        if self.config.enable_component_validation:
            log.info(f"Running compliance validation on {component}")
            compliance_validation_success = self.compliance_validation(component)
            if len(self.config.component_validation) > 0:
                log.info(f"Running customized validation on {component}")
                for jsonpath, regex in self.config.component_validation.items():
                    customized_validation_success = (
                        customized_validation_success
                        if self.customized_validation(
                            jsonpath,
                            regex,
                            component,
                            self.config.fail_if_pattern_not_found_in_component_validation,
                        )
                        else False
                    )

        if (
            validate_component_success
            and compliance_validation_success
            and customized_validation_success
        ):
            log.info(f"Component {component} is valid.")
            self.register_component_status(component, "validate", "succeeded")
        else:
            self.register_component_status(component, "validate", "failed")
            self.register_error(f"Error when validating component {component}.")

    def compliance_validation(self, component: str) -> bool:
        """
        This function checks whether a given component spec YAML file
        meets all the requirements for running in the compliant AML.
        Specifically, it checks (1) whether the image URL is compliant；
        （2）whether the pip index-url is compliant; (3) whether
        "default" is only Conda channel
        """
        with open(component, "r") as spec_file:
            spec = YAML(typ="safe").load(spec_file)

        # Check whether the docker image URL is compliant
        image_url = jsonpath_ng.parse("$.environment.docker.image").find(spec)
        if len(image_url) > 0:
            if (
                urlparse(image_url[0].value).path.split("/")[0]
                not in ALLOWED_CONTAINER_REGISTRIES
            ):
                log.error(
                    f"The container base image in {component} is not allowed for compliant run."
                )
                return False

        # check whether the package feed is compliant
        component_name = get_component_name(spec, component)
        (
            package_dependencies,
            conda_dependencies,
            conda_channels,
        ) = self._extract_dependencies_and_channels(component=component, component_name=component_name)
        if len(package_dependencies) > 0:
            for dependency in package_dependencies:
                if re.match("^--index-url", dependency) or re.match(
                    "^--extra-index-url", dependency
                ):
                    if dependency.split(" ")[1] not in ALLOWED_PACKAGE_FEEDS:
                        log.error(
                            f"The package feed in {component} is not allowed for compliant run."
                        )
                        return False
            found_package_feed = False
            for feed in ALLOWED_PACKAGE_FEEDS:
                if (
                    f"--index-url {feed}" in package_dependencies
                    or f"--extra-index-url {feed}"
                    in package_dependencies
                ):
                    found_package_feed = True
            if not found_package_feed:
                log.error(
                    f"Neither Polymer package feed nor O365 feed is found in environment of {component}"
                )
                return False
        # Check whether "default" is only Conda channel
        if len(conda_channels) > 1 or (
            len(conda_channels) == 1 and conda_channels[0] != "."
        ):
            log.error("Only the default conda channel is allowed for compliant run.")
            return False

        return True

    @staticmethod
    def customized_validation(
        jsonpath: str,
        regex: str,
        component: str,
        fail_if_pattern_not_found_in_component_validation: bool,
    ) -> bool:
        """
        This function leverages regular expressionm atching and
        JSONPath expression to enforce user-provided "strict"
        validation on Azure ML components
        """
        with open(component, "r") as spec_file:
            spec = YAML(typ="safe").load(spec_file)

        parsed_patterns = jsonpath_ng.parse(jsonpath).find(spec)
        validation_success = True
        if not parsed_patterns:
            log.warning(f"The pattern {jsonpath} is not found in {component}")
            if fail_if_pattern_not_found_in_component_validation:
                validation_success = False
        if len(parsed_patterns) > 0:
            for parsed_pattern in parsed_patterns:
                if not re.match(regex, parsed_pattern.value):
                    log.error(
                        f"The parsed pattern {parsed_pattern} in {component} doesn't match the regular expression {regex}"
                    )
                    validation_success = False
        return validation_success

    def add_conda_to_system_path(self) -> None:
        """
        This function will add the conda directory to the system path,
        if the `prepare` command is run from a GitHub Action or from an
        Azure DevOps build. It supports both Windows and Linux operating
        systems.
        """
        # let's figure out the OS
        operating_system = self.determine_operating_system()
        # let's figure out the runtime environment
        runtime_environment = self.determine_runtime_environment()
        # let's use those values to add the conda directory to the path in the appropriate manner
        self.add_conda_to_path(operating_system, runtime_environment)

    def azure_devops_append_to_path(self, value: str):
        """
        Azure DevOps doesn't support a built-in command for appending values
        to the `PATH` environment variable. It's dangerous to _prepend_ Conda
        to the path, since the current Python interpreter may not be Conda-managed.
        For more context, see:
        - https://stackoverflow.microsoft.com/questions/311834
        - https://stackoverflow.com/a/57257864
        """
        current_path = os.environ["PATH"]
        print(f"##vso[task.setvariable variable=PATH]{current_path}:{value}")

    def add_conda_to_path(
        self, 
        operating_system: OperatingSystem, 
        runtime_environment: RuntimeEnvironment, 
    ) -> None:
        """
        This function will add the conda directory to the system path,
        based on the 'operating_system' and 'runtime_environment'.
        For more details, see the following links:
        https://docs.microsoft.com/en-us/azure/devops/pipelines/ecosystems/anaconda?view=azure-devops&tabs=ubuntu#add-conda-to-your-system-path
        and https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#adding-a-system-path
        """
        success = False
        path = os.environ["PATH"]
        if operating_system == OperatingSystem.WINDOWS:
            if runtime_environment == RuntimeEnvironment.GITHUB_ACTION:
                Path(os.environ["GITHUB_PATH"]).write_text(
                    os.path.join(os.environ["CONDA"], "condabin")
                )
                success = True
            elif runtime_environment == RuntimeEnvironment.AZURE_DEVOPS_BUILD:
                self.azure_devops_append_to_path(
                    os.path.join(os.environ["CONDA"], "Scripts")
                )
                success = True
            else:
                log.info(f"Unsupported runtime environment '{runtime_environment}'.")
        elif operating_system == OperatingSystem.LINUX:
            if runtime_environment == RuntimeEnvironment.GITHUB_ACTION:
                Path(os.environ["GITHUB_PATH"]).write_text(
                    os.path.join(os.environ["CONDA"], "bin")
                )
                success = True
            elif runtime_environment == RuntimeEnvironment.AZURE_DEVOPS_BUILD:
                self.azure_devops_append_to_path(
                    os.path.join(os.environ["CONDA"], "bin")
                )
                success = True
            else:
                log.info(f"Unsupported runtime environment '{runtime_environment}'.")
        else:
            log.info(f"Unsupported operating system '{operating_system}'.")

        if success:
            log.info(
                f"Successfully added conda to the system path for runtime environment '{runtime_environment}' and operating system '{operating_system}'."
            )
        elif runtime_environment == RuntimeEnvironment.OTHER:
            log.info(f"Assuming local run for runtime environment {runtime_environment} and operating system {operating_system}. "
                     "Skipping adding conda to the system path."
            )
        else:
            log.error(
                f"Failed to add conda to the system path for runtime environment '{runtime_environment}' and operating system '{operating_system}'."
            )

    @staticmethod
    def determine_operating_system() -> OperatingSystem:
        """
        This function will determine the operating system of the
        current machine ('Windows', 'Linux', or 'Other').
        """
        if sys.platform.startswith("win"):
            return OperatingSystem.WINDOWS
        elif sys.platform.startswith("linux"):
            return OperatingSystem.LINUX
        else:
            return OperatingSystem.OTHER

    def determine_runtime_environment(self) -> RuntimeEnvironment:
        """
        This function will determine the runtime environment of the
        current machine ('GitHub Action', 'Azure DevOps Build', or 'Other').
        """
        github_ev = self.get_specific_ev(target="github")
        azure_ev = self.get_specific_ev(target="azure")
        runtime_environment = (
            self.determine_runtime_environment_based_on_environment_variables(
                github_ev, azure_ev
            )
        )
        return runtime_environment

    @staticmethod
    def get_specific_ev(target: str) -> Union[str, None]:
        """
        This function will return the contents of an environment variable
        relevant to the 'target' runtime environment.
        """
        if target == "github":
            return os.environ.get("GITHUB_REPOSITORY")
        elif target == "azure":
            return os.environ.get("BUILD_BUILDID")
        else:
            return ""

    @staticmethod
    def determine_runtime_environment_based_on_environment_variables(
        github_ev: Union[str, None], azure_ev: Union[str, None]
    ) -> RuntimeEnvironment:
        """
        This function will use the values of the runtime environment-specific
        environment variables 'github_ev' and 'azure_ev' to determine the
        runtime environment of the current machine ('GitHub Action',
        'Azure DevOps Build', or 'Other').
        """
        if github_ev and azure_ev:
            raise ValueError(
                "It seems like both the GitHub- and Azure-specific environment variables are non-empty. This shouldn't be possible and implies shrike should be using alternate environment variables to determine the runtime environment. Please open an issue in the GitHub shrike repository: https://github.com/ai-platform-ml-platform/shrike/issues."
            )
        elif github_ev:
            return RuntimeEnvironment.GITHUB_ACTION
        elif azure_ev:
            return RuntimeEnvironment.AZURE_DEVOPS_BUILD
        else:
            return RuntimeEnvironment.OTHER


if __name__ == "__main__":
    Prepare().run()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import json
import os
import hashlib
import logging
from pathlib import Path

from opencensus.ext.azure.log_exporter import AzureLogHandler

log = logging.getLogger(__name__)


def create_catalog_stub():
    """
    Function that creates a json stub of the form: {'HashAlgorithm': 'SHA256', 'CatalogItems': {}}.
    """
    json_stub = {}
    json_stub["HashAlgorithm"] = "SHA256"
    json_stub["CatalogItems"] = {}
    return json_stub


def create_SHA_256_hash_of_file(file):
    """
    Function that returns the SHA 256 hash of 'file'.\n
    Logic taken from https://www.quickprogrammingtips.com/python/how-to-calculate-sha256-hash-of-a-file-in-python.html
    """
    sha256_hash = hashlib.sha256()
    with open(file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
        # Converting to upper case because that's what is required by the policy
        # service. See their code:
        # https://dev.azure.com/msasg/Bing_and_IPG/_git/Aether?path=/src/aether/platform/backendV2/BlueBox/PolicyService/Microsoft.MachineLearning.PolicyService/Workers/CatalogValidation.cs
        return sha256_hash.hexdigest().upper()


def add_file_to_catalog(file_for_catalog, catalog, absolute_path_to_remove):
    """
    Function that adds an entry for 'file_for_catalog' to the 'catalog'.\n
    Specifically, {<Relative path of file>: <Hash of file>} will be added to the "CatalogItems" dictionary of the 'catalog' json, where <Hash of file> is computed with the create_SHA_256_hash_of_file() function, and <Relative path of file> is obtained by removing 'absolute_path_to_remove' from the full 'file_for_catalog' path
    """
    hash_of_file = create_SHA_256_hash_of_file(file_for_catalog)
    relative_path = file_for_catalog.split(absolute_path_to_remove)[1]
    catalog["CatalogItems"][relative_path] = hash_of_file
    return catalog


def write_two_catalog_files(catalog, path):
    """
    Function that writes 'catalog' into 2 duplicate files: "path/config.json" and "path/config.json.sig".
    """
    with open(os.path.join(path, "catalog.json"), "w") as jsonFile1:
        json.dump(catalog, jsonFile1)
    with open(os.path.join(path, "catalog.json.sig"), "w") as jsonFile2:
        json.dump(catalog, jsonFile2)


def delete_two_catalog_files(path):
    """
    Function that deletes the "catalog.json" and "catalog.json.sig" files located at 'path', if they exist
    """
    # catalog.json
    file_path_json = os.path.join(path, "catalog.json")
    if os.path.exists(file_path_json):
        log.warning(f"{file_path_json} already exists. Deleting it")
        os.remove(file_path_json)
    # catalog.json.sig
    file_path_json_sig = os.path.join(path, "catalog.json.sig")
    if os.path.exists(file_path_json_sig):
        log.warning(f"{file_path_json_sig} already exists. Deleting it")
        os.remove(file_path_json_sig)


class TelemetryLogger:
    """Utils class for opencensus azure monitor"""

    def __init__(
        self, enable_telemetry=True, instrumentation_key=None, level=logging.INFO
    ):
        self.logger = logging.getLogger("telemetry_logger")
        self.logger.setLevel(level)
        self.enable_telemetry = enable_telemetry
        # Why is it okay to include this key directly in the source code?
        # For any client-side tool, there is a fundamental problem with protecting instrumentation
        # keys. You want the published tool to be able to collect telemetry, but the only way
        # it can do this is if it has some kind of instrumentation key.
        #
        # For an authoritative example, the dotnet CLI contains their telemetry key in a
        # public GitHub repository:
        # https://github.com/dotnet/cli/blob/master/src/dotnet/Telemetry/Telemetry.cs
        #
        # The underlying Azure resource is called `aml1p-ml-tooling`.
        self.instrumentation_key = (
            "aaefce9e-d109-4fac-bb9f-8277c68e91ac"
            if instrumentation_key is None
            else instrumentation_key
        )
        handler = AzureLogHandler(
                connection_string=f"InstrumentationKey={self.instrumentation_key}"
            )
        handler.add_telemetry_processor(self.scrubber_function)
        self.logger.addHandler(handler)

    def log_trace(self, message, properties={}, level=logging.INFO):
        if self.enable_telemetry:
            try:
                if level == logging.INFO:
                    self.logger.info(message, extra=properties)
                elif level == logging.WARNING:
                    self.logger.warning(message, extra=properties)
                elif level == logging.ERROR:
                    self.logger.error(message, extra=properties)
                elif level == logging.CRITICAL:
                    self.logger.critical(message, extra=properties)
                else:
                    log.error("The logging level is not expected!")
            except Exception as ex:
                log.warning("Send telemetry exception: %s", str(ex))
        else:
            log.info(
                "Sending trace log messages to application insight has been disabled."
            )
    
    # Callback function to scrub some columns
    def scrubber_function(self, envelope):
        envelope.tags['ai.cloud.roleInstance'] = 'cloud_RoleInstance_Scrubbed'
        envelope.tags["ai.location.ip"] = "IP_Scrubbed"


def get_component_name(spec, source_path):
    """
    Get component name from spec (dict) or source path. If spec doesn't have name, use the name of base directory.
    """
    # default name is the name of flow directory for promptflow component
    component_name = spec.get("name", Path(source_path).parent.name)
    return component_name



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

""" Configuration dataclasses for AMLPipelineHelper """
from dataclasses import dataclass, field
from omegaconf import MISSING
from typing import Optional, Any, Dict, List
from shrike.pipeline.module_helper_base import module_loader_config, module_manifest

# Default config for HDI components
HDI_DEFAULT_CONF = '{"spark.yarn.appMasterEnv.DOTNET_ASSEMBLY_SEARCH_PATHS":"./udfs","spark.yarn.maxAppAttempts":"1","spark.yarn.appMasterEnv.PYSPARK_PYTHON":"/usr/bin/anaconda/envs/py37/bin/python3","spark.yarn.appMasterEnv.PYSPARK_DRIVER_PYTHON":"/usr/bin/anaconda/envs/py37/bin/python3"}'


@dataclass
class pipeline_cli_config:  # pylint: disable=invalid-name
    """
    Pipeline config for command line parameters

    Parameters:
    - `regenerate_outputs` (bool): set to True for forcing the re-computation of outputs, even if code or data did not change (default to False)
    - `continue_on_failure` (bool): set to True for, in the event of a step failure, forcing the not-downstream steps to keep running (default to False)
    - `disable_telemetry` (bool): set to True to disable telemetry (default to False)
    - `verbose` (bool): set to True for verbose logging (default to False)
    - `submit` (bool): set to True for actually submitting the pipeline (default to False)
    - `resume` (bool): set to True for resuming a previous pipeline run (default to False), in which case you'll also need to provide the experiment name (`experiment_name`) and the run Id (`pipeline_run_id`)
    - `canary` (bool): set to True for running a canary test after experiment completion (default to False); the *canary()* method, stubbed in the AMLPipelineHelper class but whose implementation is left to the user, can be used to check for outputs, metrics, etc...
    - `export` (Optional[str]): provide a file name if you want to export the pipeline graph (no name by default, which implies no export)
    - `silent` (bool): set to True for not automatically opening a browser window upon pipeline submission (default to False)
    - `wait` (bool): set to True to wait for the completion of the pipeline execution and show the full logs in the meantime (default to False)
    - `experiment_name` (str): to set the name of the experiment in the Azure ML portal
    - `experiment_description` (Optional[str]): to set the description of the experiment in the Azure ML portal
    - `display_name` (Optional[str]): to set the display name of the experiment in the Azure ML portal
    - `pipeline_run_id` (str): the ID of the pipeline to reuse when resume is set to True
    - `tags` (Optional[Any]): any optional tags to add to the experiment
    - `config_dir` (Optional[str]): the directory where the pipeline configuration is stored
    - `publish` (bool): set to True to publish the pipeline to Azure ML as an endpoint (default to False)
    - `endpoint_name` (Optional[str]): name of the endpoint in which to publish the pipeline
    - `endpoint_description` (Optional[str]): description of the endpoint in which to publish the pipeline
    - `log_error_only` （bool): set to True for logging error only (default to False)
    - `skip_validation` (bool): set to True to not run pipeline.validate() after building and submit with "skip_validation=True" (default to False)
    - `script` (Optional[str]): for Ray actor orchestration only, the script containing classes to be wrapped as Ray actors
    """

    regenerate_outputs: bool = False
    continue_on_failure: bool = False
    disable_telemetry: bool = False
    verbose: bool = False
    submit: bool = False
    resume: bool = False
    canary: bool = False
    export: Optional[str] = None
    silent: bool = False
    wait: bool = False
    experiment_name: str = MISSING
    experiment_description: Optional[str] = None
    display_name: Optional[str] = None
    pipeline_run_id: str = MISSING
    tags: Optional[Any] = None
    config_dir: Optional[str] = None
    publish: bool = False
    endpoint_name: Optional[str] = None
    endpoint_description: Optional[str] = None
    log_error_only: bool = False
    skip_validation: bool = False
    script: Optional[str] = None
    skip_update_dc: bool = False


@dataclass
class aml_connection_config:  # pylint: disable=invalid-name
    """
    AML connection configuration

    Parameters:
    - `subscription_id` (str): Azure subscription id
    - `resource_group` (str): Azure resource group
    - `workspace_name` (str): Azure ML workspace name
    - `tenant` (Optional[str]): Azure Active Directory tenant
    - `auth` (str): the authentication method to connect to Azure ("msi", "azurecli", or "interactive" by default).
    - `force` (bool): to force tenant authentication (False by default)
    """

    subscription_id: str = MISSING
    resource_group: str = MISSING
    workspace_name: str = MISSING
    tenant: Optional[str] = None
    auth: str = "interactive"
    force: bool = False


@dataclass
class pipeline_compute_config:  # pylint: disable=invalid-name
    """
    AML workspace compute targets and I/O modes

    Parameters:
    - `default_compute_target (str): name of default compute target to use if not specified
    - `linux_cpu_dc_target` (str): name of linux cpu detonation chamber compute target
    - `linux_cpu_prod_target` (str): name of linux cpu regular compute target
    - `linux_gpu_dc_target` (str): name of linux gpu detonation chamber compute target
    - `linux_gpu_prod_target` (str): name of linux gpu regular compute target
    - `linux_input_mode` (str): input mode for linux compute targets (default to "mount"), more details at https://componentsdk.azurewebsites.net/concepts/inputs-and-outputs.html#input-dataset-mode
    - `linux_output_mode` (str): output mode for linux compute targets (default to "mount"), more details at https://componentsdk.azurewebsites.net/concepts/inputs-and-outputs.html#output-dataset-mode
    - `windows_cpu_prod_target` (str): name of windows cpu regular compute target
    - `windows_cpu_dc_target` (str): name of windows cpu detonation chamber compute target
    - `windows_input_mode` (str): input mode for windows compute targets (default to "download"), more details at https://componentsdk.azurewebsites.net/concepts/inputs-and-outputs.html#input-dataset-mode
    - `windows_output_mode` (str): output mode for windows compute targets (default to "upload"), more details at https://componentsdk.azurewebsites.net/concepts/inputs-and-outputs.html#output-dataset-mode
    - `hdi_prod_target` (str): name of HDI regular compute target
    - `hdi_dc_target` (Optional[str]): name of HDI detonation chamber compute target
    - `hdi_driver_memory` (str): HDI driver memory in GB (default to "4g")
    - `hdi_driver_cores` (int): number of HDI driver cores (default to 2)
    - `hdi_executor_memory` (str): HDI executor memory in GB (default to "3g")
    - `hdi_executor_cores` (int): number of HDI executor cores (default to 2)
    - `hdi_number_executors` (int): number of HDI executors (default to 1)
    - `hdi_conf` (Optional[Any]): additional HDI parameters
    - `synapse_prod_target` (str): name of Synapse regular compute target
    - `synapse_dc_target` (str): name of Synapse detonation chamber compute target
    - `synapse_driver_memory` (Optional[str]): Synapse driver memory in GB
    - `synapse_driver_cores` (Optional[int]): number of Synapse driver cores
    - `synapse_executor_memory` (Optional[str]): Synapse executor memory in GB
    - `synapse_executor_cores` (Optional[int]): number of Synapse executor cores
    - `synapse_number_executors` (Optional[int]): number of Synapse executors
    - `synapse_conf` (Optional[Any]): additional Synapse parameters
    - `parallel_node_count` (int): number of nodes for parallel steps (default to 10)
    - `parallel_process_count_per_node` (Optional[int]): deprecated. Please use `parallel_process_count_per_instance`
    - `parallel_process_count_per_instance` (Optional[int]): number of processes for each node for parallel steps
    - `parallel_run_invocation_timeout` (int): threshold in seconds for parallel step invocations to timeout (default to 10800)
    - `parallel_run_max_try` (int): maximum number of times to retry parallel step invocations (default to 3)
    - `parallel_mini_batch_size` (int): mini batch size for parallel steps (default to 1)
    - `parallel_error_threshold` (int): error threshold for parallel steps (default to -1)
    - `datatransfer_target` (Optional[str]): name of DataFactory for datatransfer steps
    - `compliant_datastore` (str): name of the default compliant datastore
    - `noncompliant_datastore` (Optional[str]): name of the default non-compliant datastore
    - `dc_datastore` (Optional[str]): name of the default datastore for HDI detonation chamber and downstream steps
    """

    default_compute_target: str = MISSING
    linux_cpu_dc_target: str = MISSING
    linux_cpu_prod_target: str = MISSING
    linux_gpu_dc_target: str = MISSING
    linux_gpu_prod_target: str = MISSING
    linux_input_mode: str = "mount"
    linux_output_mode: str = "mount"

    windows_cpu_prod_target: str = MISSING
    windows_cpu_dc_target: str = MISSING
    windows_input_mode: str = "download"
    windows_output_mode: str = "upload"

    hdi_prod_target: str = MISSING
    hdi_dc_target: Optional[str] = None
    hdi_driver_memory: str = "4g"
    hdi_driver_cores: int = 2
    hdi_executor_memory: str = "3g"
    hdi_executor_cores: int = 2
    hdi_number_executors: int = 10
    hdi_conf: Optional[Any] = MISSING

    synapse_prod_target: Optional[str] = None
    synapse_dc_target: Optional[str] = None
    synapse_driver_memory: Optional[str] = MISSING
    synapse_driver_cores: Optional[int] = MISSING
    synapse_executor_memory: Optional[str] = MISSING
    synapse_executor_cores: Optional[int] = MISSING
    synapse_number_executors: Optional[int] = MISSING
    synapse_conf: Optional[Any] = MISSING

    parallel_node_count: int = 10
    parallel_process_count_per_node: Optional[int] = MISSING
    parallel_run_invocation_timeout: int = 10800
    parallel_run_max_try: int = 3
    parallel_mini_batch_size: int = 1
    parallel_error_threshold: int = -1

    datatransfer_target: Optional[str] = MISSING

    default_datastore: str = MISSING
    compliant_datastore: str = MISSING
    noncompliant_datastore: Optional[str] = MISSING
    dc_datastore: Optional[str] = None


@dataclass
class tenant_override_config:
    """
    Tenant override configuration

    Parameters:
    - `allow_override` (bool): set to True to control whether the submission-time override functionality will be executed or not (default to False)
    - `keep_modified_files` (bool): set to True for the modified files (`spec.yaml`, `env.yaml`, etc...) to be saved and renamed as `<filename>_<tenant_id>.<extension>` (default to False)
    - `mapping` (Dict[str, Any]): (nested) dictionary-style definition. If this tenant is being used with `allow_override` = True, then all **local** components will be scanned and the matching fields defined in this `mapping` section will be changed.
      - Keys: `tenant_id` (e.g.: `72f988bf-86f1-41af-91ab-2d7cd011db47`) or "aml configuration" filename in `<config-dir>/aml` (which is also used as in `defaults: aml` in this yaml file).
      - Values: (nested) dictionaries, e.g. `environment.docker.image`. You could define the override for any field in [component schema](https://componentsdk.azurewebsites.net/components.html).
      - For string-type fields such as `environment.docker.image`, the override pattern is "old_value: new_value". For dict-type fields such as `tags`, the pattern is "key: new_value".
      - See https://shrike-docs.com/pipeline/submission-time-override/ for more information
    """

    allow_override: bool = False
    keep_modified_files: bool = False
    mapping: Dict[str, Any] = field(default_factory=lambda: {})


@dataclass
class silo:
    """
    Silo info for federated learning

    Parameters:
    - `compute` (str): the compute target for this silo
    - `datastore` (str): the datastore for this silo
    - `params` (Optional[Dict[str, Any]]): additional parameters relevant to the job that will run in the silo, such as `dataset`, etc...
    - `inherit` (Optional[List[str]]): list of `config_group`s to apply to this silo, and the override priority is per-silo config > `inherit` > `default_config`
    - `location` (Optional[str]): location of the silo sandbox, required for unified workspace
    - `environment` (Optional[str]): environment type of the silo sandbox, required for unified workspace. Acceptable values include "msit", "dnd", "mdp", and "cloverport"
    """

    compute: Optional[str] = None
    datastore: Optional[str] = None
    params: Optional[Dict[str, Any]] = MISSING
    inherit: Optional[List[str]] = MISSING
    location: Optional[str] = MISSING
    environment: Optional[str] = MISSING


@dataclass
class federated_config:
    """
    Federated learning configuration

    Parameters:
    - `silos` (Dict[str, silo]): the various silos to use for federated learning
    - `orchestrator` (str): the silo that acts as the orchestrator, where aggregation happens
    - `max_iterations` (int): number of training rounds (default to 1)
    - `params` (Optional[Dict[str, Any]]): additional parameters such as `agg_weight`, `model_name`, etc...
    - `config_group` (Optional[Dict[str, Any]]): config applying to all or some silos; `default_config` will be applied to all silos, and you can also define any customized config
    - `data_transfer_component` (str): name of the data transfer component **registered** in your workspace
    - `deactivate_data_transfer` (bool): set to True to not move the data between central datastore and silo datastore, so that previous results can be reused (default to False)
    - `use_secure_aggregation` (bool): set to True when users want to use secure aggation in their federated learning pipeline (default to False)
    """

    silos: Dict[str, silo] = MISSING
    orchestrator: str = MISSING
    max_iterations: int = 1
    params: Optional[Dict[str, Any]] = None
    config_group: Optional[Dict[str, Any]] = None
    data_transfer_component: Optional[str] = MISSING
    deactivate_data_transfer: bool = False
    use_secure_aggregation: bool = False


@dataclass
class compute_config:
    """
    AML compute and info

    Parameters:
    - `name` (str): name of compute target.
    - `os` (str): os of CPU/GPU compute target, "linux" or "windows".
    - `DC` (bool): whether the compute target is detonation or not.
    - `gpu` (bool): whether the compute target is GPU or not.
    - `location` (str): location of the child `silo` where the compute target belongs, defined by user and used to identify `silo`.
        The location should be the same for all computes from one child workspace; for example,
    you may use "europe" while the GPU and CPU might be provisioned in "westeurope" and "northeurope", respectively.
    - `environment` (str): environment type of the compute target belongs, defined by user and used to identify `silo`.
    """

    name: str = MISSING
    os: str = "linux"
    DC: bool = field(default=False)
    gpu: bool = field(default=False)
    location: str = MISSING
    type: str = "amlcompute"
    environment: str = MISSING
    # TODO: what should be default values?


@dataclass
class datastore_config:
    """
    AML datastore and info

    Parameters:
    - `name` (str): name of datastore
    - `DC` (bool): whether the datastore is detonation or not
    - `location` (str): location of the child `silo` where the datastore belongs, defined by user and used to identify `silo`.
    - `environment` (str): environment type of the child `silo` where the datastore belongs, defined by user and used to identify `silo`.
    """

    name: str = MISSING
    DC: Optional[bool] = False
    location: str = MISSING
    environment: str = MISSING


@dataclass
class uw_compute_config:
    """
    Unified AML workspace compute targets and I/O modes

    Parameters:
    - `default_compute_target` (str): name of default compute target to use if not specified
    - `default_datastore` (str): name of default datastore to use if not specified
    - `computes (List)`: list of `compute_config`
    - `datastores (List)`: list of `datastore_config`
    - `linux_input_mode` (str): input mode for linux compute targets (default to "mount"), more details at https://componentsdk.azurewebsites.net/concepts/inputs-and-outputs.html#input-dataset-mode
    - `linux_output_mode` (str): output mode for linux compute targets (default to "mount"), more details at https://componentsdk.azurewebsites.net/concepts/inputs-and-outputs.html#output-dataset-mode
    - `windows_input_mode` (str): input mode for windows compute targets (default to "download"), more details at https://componentsdk.azurewebsites.net/concepts/inputs-and-outputs.html#input-dataset-mode
    - `windows_output_mode` (str): output mode for windows compute targets (default to "upload"), more details at https://componentsdk.azurewebsites.net/concepts/inputs-and-outputs.html#output-dataset-mode
    - `hdi_driver_memory` (str): HDI driver memory in GB (default to "4g")
    - `hdi_driver_cores` (int): number of HDI driver cores (default to 2)
    - `hdi_executor_memory` (str): HDI executor memory in GB (default to "3g")
    - `hdi_executor_cores` (int): number of HDI executor cores (default to 2)
    - `hdi_number_executors` (int): number of HDI executors (default to 1)
    - `hdi_conf` (Optional[Any]): additional HDI parameters
    - `synapse_driver_memory` (Optional[str]): Synapse driver memory in GB
    - `synapse_driver_cores` (Optional[int]): number of Synapse driver cores
    - `synapse_executor_memory` (Optional[str]): Synapse executor memory in GB
    - `synapse_executor_cores` (Optional[int]): number of Synapse executor cores
    - `synapse_number_executors` (Optional[int]): number of Synapse executors
    - `synapse_conf` (Optional[Any]): additional Synapse parameters
    - `parallel_node_count` (int): number of nodes for parallel steps (default to 10)
    - `parallel_process_count_per_node` (Optional[int]): deprecated. Please use `parallel_process_count_per_instance`
    - `parallel_process_count_per_instance` (Optional[int]): number of processes for each node for parallel steps
    - `parallel_run_invocation_timeout` (int): threshold in seconds for parallel step invocations to timeout (default to 10800)
    - `parallel_run_max_try` (int): maximum number of times to retry parallel step invocations (default to 3)
    - `parallel_mini_batch_size` (int): mini batch size for parallel steps (default to 1)
    - `parallel_error_threshold` (int): error threshold for parallel steps (default to -1)
    - `compliant_datastore` (str): name of the default compliant datastore
    - `noncompliant_datastore` (Optional[str]): name of the default non-compliant datastore
    - `dc_datastore` (Optional[str]): name of the default datastore for HDI detonation chamber and downstream steps
    """

    default_compute_target: str = MISSING
    default_datastore: str = MISSING
    computes: List[compute_config] = field(default_factory=list)
    datastores: List[datastore_config] = field(default_factory=list)
    noncompliant_datastore: Optional[str] = MISSING

    linux_input_mode: str = "mount"
    linux_output_mode: str = "mount"

    windows_input_mode: str = "download"
    windows_output_mode: str = "upload"

    hdi_driver_memory: str = "4g"
    hdi_driver_cores: int = 2
    hdi_executor_memory: str = "3g"
    hdi_executor_cores: int = 2
    hdi_number_executors: int = 10
    hdi_conf: Optional[Any] = MISSING

    synapse_driver_memory: Optional[str] = MISSING
    synapse_driver_cores: Optional[int] = MISSING
    synapse_executor_memory: Optional[str] = MISSING
    synapse_executor_cores: Optional[int] = MISSING
    synapse_number_executors: Optional[int] = MISSING
    synapse_conf: Optional[Any] = MISSING

    parallel_node_count: int = 10
    parallel_process_count_per_node: Optional[int] = MISSING
    parallel_run_invocation_timeout: int = 10800
    parallel_run_max_try: int = 3
    parallel_mini_batch_size: int = 1
    parallel_error_threshold: int = -1


def default_config_dict():
    """Constructs the config dictionary for the pipeline helper settings"""
    return {
        "aml": aml_connection_config,
        "run": pipeline_cli_config,
        "compute": pipeline_compute_config,
        "module_loader": module_loader_config,
        "modules": module_manifest,
        "tenant_overrides": tenant_override_config,
        "federated_config": federated_config,
        "uw_config": uw_compute_config,
    }



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
"""
Python utilities that help users to manage, validate
and submit AML pipelines in dpv2
"""

from .pipeline_helper import AMLPipelineHelper
from .module_helper import AMLModuleLoader
from .aml_connect import current_workspace

__all__ = ["AMLPipelineHelper", "AMLModuleLoader", "current_workspace"]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
"""
Avoid breaking changes to `from shrike.pipeline.pipeline_helper import AMLPipelineHelper`
"""

from .v1 import AMLPipelineHelper



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

EXPERIMENTAL_WARNING_MSG = (
    "EXPERIMENTAL: shrike.distributed is an incubation contrib package"
    + "that should be considered experimental and could change at any time."
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Tooling for making it easy to create Azure DevOps build pipelines for
validating, "building", signing, and registering components in eyes-off
Azure Machine Learning workspaces.
"""

try:
    from .commands import prepare, register
except ImportError as error:
    raise ImportError(f"{error.msg}. Please install using `pip install shrike[build]`.")



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
V1 Pipeline helper class to create pipelines loading modules from a flexible manifest.
"""
from __future__ import annotations
import argparse
from abc import ABC
import os
import json
import logging
import uuid
from typing import Callable, Optional
from toposort import toposort_flatten
import yaml
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from toposort import toposort_flatten, CircularDependencyError
from dataclasses import dataclass, fields, asdict
from functools import lru_cache, wraps
from itertools import groupby
import datetime

try:
    import hydra
    from hydra.core.hydra_config import HydraConfig
    from omegaconf import DictConfig, OmegaConf, MISSING
    from flatten_dict import flatten

    import azureml
    from azureml.core import Datastore
    from azureml.core import Dataset
    from azureml.pipeline.core import PipelineRun
    from azure.ml.component.component import Component, Input, Output
    from azure.ml.component.pipeline import Pipeline
    from azure.ml.component._pipeline_parameters import PipelineParameter
    from azureml.exceptions._azureml_exception import UserErrorException
except ImportError as error:
    raise ImportError(
        f"{error.msg}. Please install using `pip install shrike[pipeline]`."
    )

from shrike import __version__
from shrike.pipeline.pipeline_helper_base import AMLPipelineHelperBase
from shrike.pipeline.v1.aml_connect import azureml_connect, current_workspace
from shrike.pipeline.v1.module_helper import AMLModuleLoader
from shrike.pipeline.pipeline_config import (
    HDI_DEFAULT_CONF,
    pipeline_compute_config,
    uw_compute_config,
)

log = logging.getLogger(__name__)


class AMLPipelineHelper(AMLPipelineHelperBase, ABC):
    def __init__(self, config, module_loader=None):
        """Constructs the pipeline helper.

        Args:
            config (DictConfig): config for this object
            module_loader (AMLModuleLoader): which module loader to (re)use
        """
        if module_loader is None:
            log.info(
                f"Creating instance of AMLModuleLoader for {self.__class__.__name__}"
            )
            module_loader = AMLModuleLoader(config)
        super(AMLPipelineHelper, self).__init__(
            config=config, module_loader=module_loader
        )

    def _get_job_status(self, job):
        return job.get_status()

    def _wait_for_job_completion(self, job):
        return job.wait_for_completion(show_output=True)

    ##################################
    ### USER FACING HELPER METHODS ###
    ##################################
    @lru_cache()
    def component_load(self, component_key) -> Callable[..., "Component"]:
        """Loads one component from the manifest"""
        component_func = self.module_loader.load_module(
            component_key, self.required_modules()
        )

        @wraps(component_func)
        def wrapper(*args, **kwargs):
            component_step = component_func(*args, **kwargs)
            if not self.unified_workspace:
                # In unified workspace, we don't know which location/environment to use
                # so we will not call apply_smart_runsettings by default
                log.info(
                    f"Now try applying the smart runsettting as component post load operation after component initialize: "
                )
                self.apply_smart_runsettings(component_step)
                log.info(f"Successfully applied the smart runsettting. ")
            return component_step

        return wrapper

    @lru_cache()
    def module_load(self, module_key):
        """Loads one module from the manifest"""
        module_func = self.module_loader.load_module(
            module_key, self.required_modules()
        )

        @wraps(module_func)
        def wrapper(*args, **kwargs):
            module_step = module_func(*args, **kwargs)
            if not self.unified_workspace:
                log.info(
                    f"Now try applying the smart runsettting as component post load operation after component initialize: "
                )
                self.apply_smart_runsettings(module_step)
                log.info(f"Successfully applied the smart runsettting. ")
            return module_step

        return wrapper

    @lru_cache()
    def dataset_load(
        self,
        name,
        version="latest",
        datastore=None,
        path_on_datastore=None,
        description=None,
    ):
        """Loads a dataset by either id or name. If the workspace does not contain this dataset and path is given, create the dataset.

        Args:
            name (str): name or uuid of dataset to load
            version (str): if loading by name, used to specify version (default "latest")
            datastore (str): datastore for registering the dataset as <name>
            path_on_datastore (str): path for registering the dataset as <name>
            description (str): description of dataset to register
        """
        try:
            return self._dataset_load_by_name_or_id(name, version)
        except UserErrorException:
            log.info(
                f"Dataset {name} not found. Try registering from {path_on_datastore} on {datastore}..."
            )
            dataset = Dataset.File.from_files(
                path=[(Datastore(self.workspace(), datastore), path_on_datastore)],
                validate=False,
            )
            dataset = dataset.register(
                workspace=self.workspace(),
                name=name,
                description=description,
                create_new_version=True,
            )
            return dataset

    @lru_cache()
    def workspace(self):
        """Gets the current workspace"""
        return current_workspace()

    def connect(self):
        """Connect to the AML workspace using internal config"""
        # Only call azureml_connect if there is not an already a pre-existing workspace
        try:
            return self.workspace()
        except:
            return azureml_connect(
                aml_subscription_id=self.config.aml.subscription_id,
                aml_resource_group=self.config.aml.resource_group,
                aml_workspace_name=self.config.aml.workspace_name,
                aml_auth=self.config.aml.auth,
                aml_tenant=self.config.aml.tenant,
                aml_force=self.config.aml.force,
            )  # NOTE: this also stores aml workspace in internal global variable

    ################
    ### MAIN/RUN ###
    ################
    def apply_recommended_runsettings(
        self,
        module_name,
        module_instance,
        gpu=False,  # can't autodetect that
        hdi="auto",
        windows="auto",
        parallel="auto",
        mpi="auto",
        scope="auto",
        datatransfer="auto",
        sweep="auto",
        synapse="auto",
        location=None,
        environment=None,
        **custom_runtime_arguments,
    ):
        """Applies regular settings for a given module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            gpu (bool): is the module using GPU?
            hdi (bool): is the module using HDI/Spark?
            windows (bool): is the module using Windows compute?
            parallel (bool): is the module using ParallelRunStep?
            mpi (bool): is the module using Mpi?
            scope (bool): is the component using scope?
            datatransfer (bool): is the component using datatransfer?
            sweep (bool): is the component using sweep?
            synapse (bool): is the component using Synapse?
            location (str): location of the silo to use in unified workspace setting
            environment (str): environment of the silo to use in unified workspace setting
            custom_runtime_arguments (dict): any additional custom args
        """
        # Verifies if module_name corresponds to module_instance
        self._check_module_runsettings_consistency(module_name, module_instance)

        # Skips if this is a PipelineComponent
        if str(module_instance.type) == "PipelineComponent":
            log.info(f"Component {module_name} detected as PipelineComponent.")
            return

        if environment:
            assert environment.lower() in [
                "cloverport",
                "msit",
                "mdp",
                "dnd",
            ]  # TODO: what else allowed?

        # Auto detects runsettings
        if hdi == "auto":
            hdi = str(module_instance.type) == "HDInsightComponent"
            if hdi:
                log.info(f"Module {module_name} detected as HDI: {hdi}")

        if parallel == "auto":
            parallel = str(module_instance.type) == "ParallelComponent"
            if parallel:
                log.info(f"Module {module_name} detected as PARALLEL: {parallel}")

        if mpi == "auto":
            mpi = str(module_instance.type) == "DistributedComponent"
            if mpi:
                log.info(f"Module {module_name} detected as MPI: {mpi}")

        if scope == "auto":
            scope = str(module_instance.type) == "ScopeComponent"
            if scope:
                log.info(f"Module {module_name} detected as SCOPE: {scope}")

        if sweep == "auto":
            sweep = str(module_instance.type) == "SweepComponent"
            if sweep:
                log.info(f"Module {module_name} detected as SweepComponent: {sweep}")

        if synapse == "auto":
            synapse = str(module_instance.type) in ["SparkComponent", "spark"]
            if synapse:
                log.info(f"Module {module_name} detected as spark (Synapse): {synapse}")

        if windows == "auto":
            if (
                str(module_instance.type) == "HDInsightComponent"
                or str(module_instance.type) == "ScopeComponent"
                or str(module_instance.type) == "DataTransferComponent"
                or str(module_instance.type) == "SweepComponent"
                or str(module_instance.type) == "SparkComponent"
            ):
                # HDI/scope/datatransfer/sweep modules might not have that environment object
                windows = False
            else:
                windows = (
                    module_instance._definition.environment.os.lower() == "windows"
                )
                if windows:
                    log.info(f"Module {module_name} detected as WINDOWS: {windows}")

        if datatransfer == "auto":
            datatransfer = str(module_instance.type) == "DataTransferComponent"
            if datatransfer:
                log.info(
                    f"Module {module_name} detected as DATATRANSFER: {datatransfer}"
                )

        if parallel:
            self._apply_parallel_runsettings(
                module_name,
                module_instance,
                windows=windows,
                gpu=gpu,
                location=location,
                environment=environment,
                **custom_runtime_arguments,
            )
            return

        if sweep:
            self._apply_sweep_runsettings(
                module_name,
                module_instance,
                windows=windows,
                gpu=gpu,
                location=location,
                environment=environment,
                **custom_runtime_arguments,
            )
            return

        if windows:
            self._apply_windows_runsettings(
                module_name,
                module_instance,
                mpi=mpi,
                location=location,
                environment=environment,
                **custom_runtime_arguments,
            )
            return

        if hdi:
            self._apply_hdi_runsettings(
                module_name,
                module_instance,
                location=location,
                environment=environment,
                **custom_runtime_arguments,
            )
            return

        if scope:
            self._apply_scope_runsettings(
                module_name,
                module_instance,
                location=location,
                environment=environment,
                **custom_runtime_arguments,
            )
            return

        if datatransfer:
            self._apply_datatransfer_runsettings(
                module_name,
                module_instance,
                location=location,
                environment=environment,
                **custom_runtime_arguments,
            )
            return

        if synapse:
            self._apply_synapse_runsettings(
                module_name,
                module_instance,
                location=location,
                environment=environment,
                **custom_runtime_arguments,
            )
            return

        self._apply_linux_runsettings(
            module_name,
            module_instance,
            mpi=mpi,
            gpu=gpu,
            location=location,
            environment=environment,
            **custom_runtime_arguments,
        )

    def apply_smart_runsettings(
        self,
        component_instance,
        gpu=False,  # can't autodetect that
        hdi="auto",
        windows="auto",
        parallel="auto",
        mpi="auto",
        scope="auto",
        datatransfer="auto",
        sweep="auto",
        synapse="auto",
        location=None,
        environment=None,
        **custom_runtime_arguments,
    ):
        """Applies regular settings for a given component.

        Args:
            component_instance (Component): the AML component we need to add settings to
            gpu (bool): is the component using GPU?
            hdi (bool): is the component using HDI?
            windows (bool): is the component using Windows compute?
            parallel (bool): is the component using ParallelRunStep?
            mpi (bool): is the component using Mpi?
            scope (bool): is the component using scope?
            datatransfer (bool): is the component using datatransfer?
            sweep (bool): is the component using sweep?
            synapse (bool): is the component using Synapse?
            location (str): location of the silo to use in unified workspace setting
            environment (str): environment of the silo to use in unified workspace setting
            custom_runtime_arguments (dict): any additional custom args
        """
        # infer component_name
        component_name = self._get_component_name_from_instance(component_instance)
        self.apply_recommended_runsettings(
            component_name,
            component_instance,
            gpu,
            hdi,
            windows,
            parallel,
            mpi,
            scope,
            datatransfer,
            sweep,
            synapse,
            location,
            environment,
            **custom_runtime_arguments,
        )

    def _get_pipeline_run(self, run_id, experiment):
        return PipelineRun(experiment, self.config.run.pipeline_run_id)

    def _get_run_url(self, pipeline_run):
        return pipeline_run.get_portal_url()

    def _dataset_load_by_name_or_id(self, name, version):
        """Loads a dataset by either id or name.

        Args:
            name (str): name or uuid of dataset to load
            version (str): if loading by name, used to specify version (default "latest")

        NOTE: in AzureML SDK there are 2 different methods for loading dataset
        one for id, one for name. This method just wraps them up in one."""
        # test if given name is a uuid
        try:
            parsed_uuid = uuid.UUID(name)
            log.info(f"Getting a dataset handle [id={name}]...")
            return Dataset.get_by_id(self.workspace(), id=name)
        except ValueError:
            log.info(f"Getting a dataset handle [name={name} version={version}]...")
            return Dataset.get_by_name(self.workspace(), name=name, version=version)

    def _set_all_inputs_to(self, module_instance, input_mode):
        """Sets all module inputs to a given intput mode"""
        input_names = [
            a
            for a in dir(module_instance.inputs)
            if isinstance(getattr(module_instance.inputs, a), Input)
        ]
        for input_key in input_names:
            input_instance = getattr(module_instance.inputs, input_key)
            input_instance.configure(mode=input_mode)
            log.info(f"Configured input {input_key} to use mode {input_mode}")

    def _set_all_outputs_to(
        self, module_instance, output_mode, compliant=True, datastore_name=None
    ):
        """Sets all module outputs to a given output mode"""
        output_names = [
            a
            for a in dir(module_instance.outputs)
            if isinstance(getattr(module_instance.outputs, a), Output)
        ]
        if not datastore_name:
            datastore_name = self._get_datastore_name(module_instance, compliant)
        for output_key in output_names:
            output_instance = getattr(module_instance.outputs, output_key)
            if output_mode:
                output_instance.configure(output_mode=output_mode)
            output_instance.configure(datastore=datastore_name)
            log.info(
                f"Configured output {output_key} to use mode {output_mode} and datastore {datastore_name}"
            )

    @lru_cache()
    def _get_uw_configs(self, assets, key_func):
        assets_dict = {}
        for key, group in groupby(assets, key_func):
            for x in group:
                pass
            assets_dict[key] = x["name"]  # list(x["name"] for x in group)
        return assets_dict

    @lru_cache()
    def _get_uw_datastores(self, location: str, environment: str, dc: bool):
        environment = environment.lower()
        if environment == "cloverport":
            # no DC concept for eyes-on workspace
            key_func = lambda x: (x["location"].lower(), x["environment"].lower())
            return self._get_uw_configs(self.config.uw_config.datastores, key_func)[
                location, environment
            ]
        else:
            key_func = lambda x: (
                x["location"].lower(),
                x["environment"].lower(),
                x["DC"],
            )
            return self._get_uw_configs(self.config.uw_config.datastores, key_func)[
                location, environment, dc
            ]

    @lru_cache()
    def _get_uw_computes(
        self,
        location: str,
        environment: str,
        type: str,
        os: str = None,
        gpu: bool = None,
        dc: bool = None,
    ):
        location = location.lower()
        environment = environment.lower()
        type = type.lower()
        if type == "datafactory":
            key_func = lambda x: (
                x["location"].lower(),
                x["environment"].lower(),
                x["type"].lower(),
            )
        elif environment == "cloverport":
            key_func = lambda x: (
                x["location"].lower(),
                x["environment"].lower(),
                x["type"].lower(),
                x["os"].lower(),
                x["gpu"],
            )
        else:
            key_func = lambda x: (
                x["location"].lower(),
                x["environment"].lower(),
                x["type"].lower(),
                x["os"].lower(),
                x["gpu"],
                x["DC"],
            )
        uw_computes = self._get_uw_configs(self.config.uw_config.computes, key_func)
        if type == "datafactory":
            return uw_computes[location, environment, type]
        elif environment == "cloverport":
            return uw_computes[location, environment, type, os.lower(), gpu]
        else:
            return uw_computes[location, environment, type, os.lower(), gpu, dc]

    def _apply_windows_runsettings(
        self,
        module_name,
        module_instance,
        mpi=False,
        target=None,
        node_count=None,
        process_count_per_node=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        location=None,
        environment=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a windows module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            mpi (bool): is job mpi ?
            target (str): force target compute over hydra conf
            node_count (int): force node_count over hydra conf
            process_count_per_node (int): force process_count_per_node over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            datastore_name (str): output datastore. This will override the datastore based on silo
            location (str): location of the silo to use in unified workspace setting
            environment (str): environment of the silo to use in unified workspace setting
            custom_runtime_arguments (dict): any additional custom args
        """
        if self.unified_workspace:
            target = target or self._get_uw_computes(
                location,
                environment,
                "amlcompute",
                "windows",
                False,
                self.module_loader.is_local(module_name),
            )
            datastore_name = self._get_uw_datastores(
                location, environment, self.module_loader.is_local(module_name)
            )  # TODO: how does hdi/spark dc work?
        else:
            if self.module_loader.is_local(module_name):
                target = (
                    target
                    if target is not None
                    else self.config.compute.windows_cpu_dc_target
                )
            else:
                target = (
                    target
                    if target is not None
                    else self.config.compute.windows_cpu_prod_target
                )

        log.info(
            f"Using windows compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        module_instance.runsettings.configure(target=target, **custom_runtime_arguments)
        if mpi:
            node_count = node_count if node_count is not None else 1
            process_count_per_node = (
                process_count_per_node if process_count_per_node is not None else 1
            )
            log.info(
                f"Using mpi with node_count={node_count} process_count_per_node={process_count_per_node}"
            )
            module_instance.runsettings.resource_layout.configure(
                node_count=node_count,
                process_count_per_node=process_count_per_node,
            )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        else:
            self._set_all_inputs_to(
                module_instance,
                self.config.compute.windows_input_mode,  # TODO: verify if this works in uw
            )

        if output_mode:
            self._set_all_outputs_to(
                module_instance, output_mode, datastore_name=datastore_name
            )
        else:
            self._set_all_outputs_to(
                module_instance,
                self.config.compute.windows_output_mode,
                datastore_name=datastore_name,
            )

    def _apply_hdi_runsettings(
        self,
        module_name,
        module_instance,
        target=None,
        driver_cores=None,
        driver_memory=None,
        executor_memory=None,
        executor_cores=None,
        number_executors=None,
        conf=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        location=None,
        environment=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a HDI module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            target (str): force target compute over hydra conf
            driver_cores (int): force driver_cores over hydra conf
            driver_memory (str): force driver_memory over hydra conf
            executor_memory (int): force executor_memory over hydra conf
            executor_cores (int): force executor_cores over hydra conf
            number_executors (int): force number_executors over hydra conf
            conf (str): force conf over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            datastore_name (str): output datastore. This will override the datastore based on silo
            location (str): location of the silo to use in unified workspace setting
            environment (str): environment of the silo to use in unified workspace setting
            custom_runtime_arguments (dict): any additional custom args
        """
        merged_conf = json.loads(HDI_DEFAULT_CONF)
        if self.unified_workspace:
            target = target or self._get_uw_computes(
                location,
                environment,
                "hdinsight",
                "linux",
                False,
                self.module_loader.is_local(module_name),
            )
            datastore_name = datastore_name or self._get_uw_datastores(
                location, environment, self.module_loader.is_local(module_name)
            )  # TODO: how does hdi/spark dc work?
        else:
            if self.module_loader.is_local(module_name):
                target = (
                    target if target is not None else self.config.compute.hdi_dc_target
                )
                if not self.config.compute.hdi_dc_target:
                    raise Exception(
                        f"Your HDI component {module_name} is using local version. Please specify hdi_dc_target"
                    )
            else:
                target = (
                    target
                    if target is not None
                    else self.config.compute.hdi_prod_target
                )
        log.info(
            f"Using HDI compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        merged_conf = json.loads(HDI_DEFAULT_CONF)
        new_conf = (
            self.config.compute.hdi_conf if "hdi_conf" in self.config.compute else None
        )
        if conf is not None:
            new_conf = conf
        if new_conf is not None:
            if isinstance(new_conf, str):
                new_conf = json.loads(new_conf)
            elif isinstance(new_conf, DictConfig):
                new_conf = flatten(dict(new_conf), reducer="dot")
            else:
                raise ValueError(
                    "computed.hdi_conf is not a valid json string or a single tested configuration."
                )
            merged_conf.update(new_conf)

        module_instance.runsettings.configure(target=target)

        module_instance.runsettings.hdinsight.configure(
            driver_memory=driver_memory
            if driver_memory is not None
            else self.config.compute.hdi_driver_memory,
            driver_cores=driver_cores
            if driver_cores is not None
            else self.config.compute.hdi_driver_cores,
            executor_memory=executor_memory
            if executor_memory is not None
            else self.config.compute.hdi_executor_memory,
            executor_cores=executor_cores
            if executor_cores is not None
            else self.config.compute.hdi_executor_cores,
            number_executors=number_executors
            if number_executors is not None
            else self.config.compute.hdi_number_executors,
            conf=merged_conf,
            **custom_runtime_arguments,
        )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(
            module_instance, output_mode, datastore_name=datastore_name
        )

    def _apply_synapse_runsettings(
        self,
        module_name,
        module_instance,
        target=None,
        driver_cores=None,
        driver_memory=None,
        executor_memory=None,
        executor_cores=None,
        number_executors=None,
        conf=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        location=None,
        environment=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a Synapse component. Shrike will not set default runsettings except for spark_identity.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            target (str): force target compute over hydra conf
            driver_cores (int): force driver_cores over hydra conf
            driver_memory (str): force driver_memory over hydra conf
            executor_memory (int): force executor_memory over hydra conf
            executor_cores (int): force executor_cores over hydra conf
            number_executors (int): force number_executors over hydra conf
            conf (str): force conf over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            datastore_name (str): output datastore. This will override the datastore based on silo
            location (str): location of the silo to use in unified workspace setting
            environment (str): environment of the silo to use in unified workspace setting
            custom_runtime_arguments (dict): any additional custom args
        """

        if self.unified_workspace:
            target = target or self._get_uw_computes(
                location,
                environment,
                "synapse",
                "linux",
                False,
                self.module_loader.is_local(module_name),
            )
            datastore_name = datastore_name or self._get_uw_datastores(
                location, environment, self.module_loader.is_local(module_name)
            )  # TODO: how does hdi/spark dc work?
        else:
            if self.module_loader.is_local(module_name):
                target = (
                    target
                    if target is not None
                    else self.config.compute.synapse_dc_target
                )
            else:
                target = (
                    target
                    if target is not None
                    else self.config.compute.synapse_prod_target
                )
        log.info(
            f"Using Synapse compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        new_conf = (
            self.config.compute.synapse_conf
            if "synapse_conf" in self.config.compute
            else None
        )
        if conf is not None:
            new_conf = conf
        if new_conf is not None:
            if isinstance(new_conf, str):
                new_conf = json.loads(new_conf)
            elif isinstance(new_conf, DictConfig) or isinstance(new_conf, dict):
                new_conf = flatten(dict(new_conf), reducer="dot")
            else:
                raise TypeError(
                    f"compute.synapse_conf of type {type(new_conf)} is not a valid json string or a single tested configuration."
                )
            if self.module_loader.is_local(module_name):
                # https://msdata.visualstudio.com/Vienna/_workitems/edit/1918995
                new_conf = json.dumps(new_conf)

            module_instance.runsettings.spark.configure(
                conf=new_conf,
            )

        module_instance.runsettings.configure(target=target)
        module_instance.runsettings.spark_identity.configure(type="managed")

        if "synapse_driver_memory" in self.config.compute:
            driver_memory = driver_memory or self.config.compute.synapse_driver_memory
        if driver_memory:
            module_instance.runsettings.spark.configure(driver_memory=driver_memory)
        if "synapse_driver_cores" in self.config.compute:
            driver_cores = driver_cores or self.config.compute.synapse_driver_cores
        if driver_cores:
            module_instance.runsettings.spark.configure(driver_cores=driver_cores)
        if "synapse_executor_memory" in self.config.compute:
            executor_memory = (
                executor_memory or self.config.compute.synapse_executor_memory
            )
        if executor_memory:
            module_instance.runsettings.spark.configure(executor_memory=executor_memory)
        if "synapse_number_executors" in self.config.compute:
            number_executors = (
                number_executors or self.config.compute.synapse_number_executors
            )
        if number_executors:
            module_instance.runsettings.spark.configure(
                number_executors=number_executors
            )
        if "synapse_executor_cores" in self.config.compute:
            executor_cores = (
                executor_cores or self.config.compute.synapse_executor_cores
            )
        if executor_cores:
            module_instance.runsettings.spark.configure(executor_cores=executor_cores)
        module_instance.runsettings.spark.configure(
            **custom_runtime_arguments,
        )
        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(
            module_instance, output_mode, datastore_name=datastore_name
        )

    def _apply_parallel_runsettings(
        self,
        module_name,
        module_instance,
        windows=False,
        gpu=False,
        target=None,
        node_count=None,
        process_count_per_node=None,
        mini_batch_size=None,
        run_invocation_timeout=None,
        run_max_try=None,
        error_threshold=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        location=None,
        environment=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a ParallelRunStep linux module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            windows (bool): is the module using Windows compute?
            gpu (bool): is the module using GPU compute?
            target (str): force target compute over hydra conf
            node_count (int): force node_count over hydra conf
            process_count_per_node (int): force process_count_per_node over hydra conf
            mini_batch_size (int): force mini_batch_size over hydra conf
            run_invocation_timeout (int): force run_invocation_timeout over hydra conf
            run_max_try (int): force run_max_try over hydra conf
            error_threshold (int): The number of file failures for the input FileDataset that should be ignored during processing.
                If the error count goes above this value, then the job will be aborted.
                Error threshold is for the entire input and not for individual mini-batches sent to run() method.
                The range is [-1, int.max]. -1 indicates ignoring all failures during processing.
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            datastore_name (str): output datastore. This will override the datastore based on silo
            location (str): location of the silo to use in unified workspace setting
            environment (str): environment of the silo to use in unified workspace setting
            custom_runtime_arguments (dict): any additional custom args
        """
        if self.unified_workspace:
            if windows and gpu:
                raise ValueError(
                    "A GPU compute target with Windows OS is not available yet!"
                )
            target = target or self._get_uw_computes(
                location,
                environment,
                "amlcompute",
                "windows" if windows else "linux",
                gpu,
                self.module_loader.is_local(module_name),
            )
            datastore_name = datastore_name or self._get_uw_datastores(
                location, environment, self.module_loader.is_local(module_name)
            )  # TODO: how does hdi/spark dc work?
        else:
            if self.module_loader.is_local(module_name):
                if windows:
                    if gpu:
                        raise ValueError(
                            "A GPU compute target with Windows OS is not available yet!"
                        )
                    else:
                        _target = self.config.compute.windows_cpu_dc_target
                else:
                    if gpu:
                        _target = self.config.compute.linux_gpu_dc_target
                    else:
                        _target = self.config.compute.linux_cpu_dc_target
            else:
                if windows:
                    if gpu:
                        raise ValueError(
                            "A GPU compute target with Windows OS is not available yet!"
                        )
                    else:
                        _target = self.config.compute.windows_cpu_prod_target
                else:
                    if gpu:
                        _target = self.config.compute.linux_gpu_prod_target
                    else:
                        _target = self.config.compute.linux_cpu_prod_target

            target = target if target is not None else _target

        log.info(
            f"Using parallelrunstep compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        module_instance.runsettings.configure(target=target)

        module_instance.runsettings.parallel.configure(
            node_count=node_count
            if node_count is not None
            else self.config.compute.parallel_node_count,
            process_count_per_node=process_count_per_node
            if process_count_per_node is not None
            else (
                self.config.compute.parallel_process_count_per_node
                if "parallel_process_count_per_node" in self.config.compute
                else None
            ),
            mini_batch_size=str(
                mini_batch_size
                if mini_batch_size is not None
                else self.config.compute.parallel_mini_batch_size
            ),
            run_invocation_timeout=run_invocation_timeout
            if run_invocation_timeout is not None
            else self.config.compute.parallel_run_invocation_timeout,
            run_max_try=run_max_try
            if run_max_try is not None
            else self.config.compute.parallel_run_max_try,
            error_threshold=error_threshold
            if error_threshold is not None
            else self.config.compute.parallel_error_threshold,
            **custom_runtime_arguments,
        )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(
            module_instance, output_mode, datastore_name=datastore_name
        )

    def _apply_linux_runsettings(
        self,
        module_name,
        module_instance,
        mpi=False,
        gpu=False,
        target=None,
        node_count=None,
        process_count_per_node=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        location=None,
        environment=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for linux module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            mpi (bool): is the job mpi?
            gpu (bool): is the job using GPU?
            target (str): force target compute over hydra conf
            node_count (int): force node_count over hydra conf
            process_count_per_node (int): force process_count_per_node over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            datastore_name (str): output datastore. This will override the datastore based on silo
            location (str): location of the silo to use in unified workspace setting
            environment (str): environment of the silo to use in unified workspace setting
            custom_runtime_arguments (dict): any additional custom args
        """
        if self.unified_workspace:
            target = target or self._get_uw_computes(
                location,
                environment,
                "amlcompute",
                "linux",
                gpu,
                self.module_loader.is_local(module_name),
            )
            datastore_name = datastore_name or self._get_uw_datastores(
                location, environment, self.module_loader.is_local(module_name)
            )  # TODO: how does hdi/spark dc work?
        else:
            if self.module_loader.is_local(module_name) and gpu:
                target = (
                    target
                    if target is not None
                    else self.config.compute.linux_gpu_dc_target
                )
            elif not self.module_loader.is_local(module_name) and gpu:
                target = (
                    target
                    if target is not None
                    else self.config.compute.linux_gpu_prod_target
                )
            elif self.module_loader.is_local(module_name) and not gpu:
                target = (
                    target
                    if target is not None
                    else self.config.compute.linux_cpu_dc_target
                )
            elif not self.module_loader.is_local(module_name) and not gpu:
                target = (
                    target
                    if target is not None
                    else self.config.compute.linux_cpu_prod_target
                )
        processor = "GPU" if gpu else "CPU"
        source = "local" if self.module_loader.is_local(module_name) else "registered"
        log.info(
            f"Using target {target} for {source} {processor} module {module_name} from pipeline class {self.__class__.__name__}"
        )

        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        module_instance.runsettings.configure(target=target, **custom_runtime_arguments)

        if mpi:
            node_count = node_count if node_count is not None else 1
            process_count_per_node = (
                process_count_per_node if process_count_per_node is not None else 1
            )
            log.info(
                f"Using mpi with node_count={node_count} process_count_per_node={process_count_per_node}"
            )
            module_instance.runsettings.resource_layout.configure(
                node_count=node_count,
                process_count_per_node=process_count_per_node,
            )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        else:
            self._set_all_inputs_to(
                module_instance, self.config.compute.linux_input_mode
            )
        if output_mode:
            self._set_all_outputs_to(
                module_instance, output_mode, datastore_name=datastore_name
            )
        else:
            self._set_all_outputs_to(
                module_instance,
                self.config.compute.linux_output_mode,
                datastore_name=datastore_name,
            )

    def _apply_scope_runsettings(
        self,
        module_name,
        module_instance,
        input_mode=None,
        output_mode=None,
        scope_param=None,
        custom_job_name_suffix=None,
        adla_account_name=None,
        priority=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a scope module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            scope_param (str): Parameters to pass to scope e.g. Nebula parameters, VC allocation parameters etc.
            custom_job_name_suffix (str): Optional parameter defining custom string to append to job name
            adla_account_name (str): The name of the Cosmos-migrated Azure Data Lake Analytics account to submit scope job
            priority (int): Scope job priority. Default is 1000.
            custom_runtime_arguments (dict): any additional custom args
        """
        log.info(
            f"Using scope compute target to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(module_instance, output_mode, compliant=False)

        module_instance.runsettings.scope.configure(
            adla_account_name=adla_account_name,
            scope_param=scope_param,
            custom_job_name_suffix=custom_job_name_suffix,
            priority=priority,
        )

    def _apply_datatransfer_runsettings(
        self,
        module_name,
        module_instance,
        compliant=True,
        target=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        location=None,
        environment=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a HDI module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            compliant (bool): destination datastore, `compliant_datastore` if True, else `noncompliant_datastore`
            target (str): force target compute over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            datastore_name (str): output datastore. This will override the datastore based on silo
            location (str): location of the silo to use in unified workspace setting
            environment (str): environment of the silo to use in unified workspace setting
            custom_runtime_arguments (dict): any additional custom args
        """
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        if self.unified_workspace:
            target = target or self._get_uw_computes(
                location, environment, "datafactory"
            )
            if compliant:
                datastore_name = datastore_name or self._get_uw_datastores(
                    location, environment, False
                )
            else:
                datastore_name = (
                    datastore_name or self.config.compute.noncompliant_datastore
                )
            self._set_all_outputs_to(
                module_instance, output_mode, datastore_name=datastore_name
            )
        else:
            target = target or self.config.compute.datatransfer_target
            self._set_all_outputs_to(
                module_instance, output_mode, compliant, datastore_name
            )
        log.info(
            f"Using datatransfer compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        module_instance.runsettings.configure(target=target)

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)

    def _apply_sweep_runsettings(
        self,
        module_name,
        module_instance,
        windows=False,
        gpu=False,
        target=None,
        input_mode=None,
        output_mode=None,
        node_count=None,
        process_count_per_node=None,
        algorithm=None,
        primary_metric=None,
        goal=None,
        policy_type=None,
        evaluation_interval=None,
        delay_evaluation=None,
        slack_factor=None,
        slack_amount=None,
        truncation_percentage=None,
        max_total_trials=None,
        max_concurrent_trials=None,
        timeout_minutes=None,
        datastore_name=None,
        location=None,
        environment=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a sweep component.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            target (str): force target compute over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf

            # For below sweep specific parameters configurations, see below doc link for more info:
            # https://componentsdk.azurewebsites.net/components/sweep_component.html#set-runsettings
            algorithm (str): sweep sampling method
            primary_metric (str): the primary metric of the hyperparameter tuning to optimize
            goal (str): whether the primary metric will be maximize or minimize when evaluating the trials
            policy_type (str): sweep early termination policy type
            evaluation_interval (int): frequency of applying the policy
            delay_evaluation (int): delays the first policy evaluation for a specified number of intervals
            slack_factor (float): the slack allowed with respect to the best performing training run, as a ratio
            slack_amount (float): the slack allowed with respect to the best performing training run, as an absolute ampunt. You should only specify either slack_factor or slack_amount, but not both.
            truncation_percentage (int): the percentage of lowest performing runs to terminate at each evaluation interval. An integer value between 1 and 99.
            max_total_trials (int): maximum number of trial runs. Must be an integer between 1 and 1000.
            max_concurrent_trials (int): maximum number of runs that can run concurrently. If not specified, all runs launch in parallel. If specified, must be an integer between 1 and 100.
            timeout_minutes (int): maximum duration, in minutes, of the hyperparameter tuning experiment. Runs after this duration are canceled.

            datastore_name (str): output datastore
            location (str): location of the silo to use in unified workspace setting
            environment (str): environment of the silo to use in unified workspace setting

            custom_runtime_arguments (dict): any additional custom args
        """
        if self.unified_workspace:
            target = target or self._get_uw_computes(
                location,
                environment,
                "amlcompute",
                "windows" if windows else "linux",
                gpu,
                self.module_loader.is_local(module_name),
            )
            datastore_name = datastore_name or self._get_uw_datastores(
                location, environment, self.module_loader.is_local(module_name)
            )  # TODO: how does hdi/spark dc work?
        else:
            if target is not None:
                target = target
            elif self.module_loader.is_local(module_name):
                if windows:
                    if gpu:
                        raise ValueError(
                            "A GPU compute target with Windows OS is not available yet!"
                        )
                    else:
                        target = self.config.compute.windows_cpu_dc_target
                else:
                    if gpu:
                        target = self.config.compute.linux_gpu_dc_target
                    else:
                        target = self.config.compute.linux_cpu_dc_target
            else:
                if windows:
                    if gpu:
                        raise ValueError(
                            "A GPU compute target with Windows OS is not available yet!"
                        )
                    else:
                        target = self.config.compute.windows_cpu_prod_target
                else:
                    if gpu:
                        target = self.config.compute.linux_gpu_prod_target
                    else:
                        target = self.config.compute.linux_cpu_prod_target
        log.info(
            f"Using target {target} to run sweep component {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        module_instance.runsettings.configure(target=target, **custom_runtime_arguments)

        if node_count:
            log.info(
                f"Setting node_count={node_count} and process_count_per_node={process_count_per_node} as run settings for sweep component {module_name} from pipeline class {self.__class__.__name__}"
            )
            module_instance.runsettings.resource_layout.configure(
                node_count=node_count, process_count_per_node=process_count_per_node
            )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(
            module_instance, output_mode, datastore_name=datastore_name
        )

        if algorithm:
            module_instance.runsettings.sweep.algorithm = algorithm
        if primary_metric:
            module_instance.runsettings.sweep.objective.configure(
                primary_metric=primary_metric
            )
        if goal:
            module_instance.runsettings.sweep.objective.configure(goal=goal)
        if policy_type:
            module_instance.runsettings.sweep.early_termination.configure(
                policy_type=policy_type
            )
        if evaluation_interval:
            module_instance.runsettings.sweep.early_termination.configure(
                evaluation_interval=evaluation_interval
            )
        if delay_evaluation:
            module_instance.runsettings.sweep.early_termination.configure(
                delay_evaluation=delay_evaluation
            )
        if slack_factor:
            module_instance.runsettings.sweep.early_termination.configure(
                slack_factor=slack_factor
            )
        if slack_amount:
            module_instance.runsettings.sweep.early_termination.configure(
                slack_amount=slack_amount
            )
        if truncation_percentage:
            module_instance.runsettings.sweep.early_termination.configure(
                truncation_percentage=truncation_percentage
            )
        if max_total_trials:
            module_instance.runsettings.sweep.limits.configure(
                max_total_trials=max_total_trials
            )
        if max_concurrent_trials:
            module_instance.runsettings.sweep.limits.configure(
                max_concurrent_trials=max_concurrent_trials
            )
        if timeout_minutes:
            module_instance.runsettings.sweep.limits.configure(
                timeout_minutes=timeout_minutes
            )

    def _get_component_name_helper(self, component_instance):
        return component_instance.name

    def _update_dc_configs(self, pipeline: Pipeline, debug: bool = False) -> Pipeline:
        # collect the known DC targets
        dc_datastore = None
        if not self.unified_workspace:
            # workaround for missing "throw_on_missing" in currently used omegaconf
            compute_config_as_dict = yaml.safe_load(
                OmegaConf.to_yaml(self.config.compute)
            )
            dc_targets: Set[str] = {
                target
                for key, target in compute_config_as_dict.items()
                if key.endswith("_dc_target") and target and target != str(MISSING)
            }
            dc_datastore = self.config.compute.dc_datastore
        else:
            log.warn(
                """
                Updating for DC is not fully supported for multi-region pipelines 
                in the unified workspace mode and is subject to error. 
                Strongly recommend to skip this by setting `run.skip_update_dc`. 
                If you need to test on DC, please try `module_loader.use_local='!<data_transfer>'`
                """
            )
            dc_targets: Set[str] = {
                target["name"]
                for target in self.config.uw_config.computes
                if target["DC"]
            }
        if not dc_targets:
            # This is kind of pointless because if one local component exists
            # then one of the `apply_*` methods wil fail if no dc target is not
            # specified. And if no local component exists, then there is no
            # point in outputting this. But let's do it anyway:
            log.info("The compute config does not contain any DC targets.")
            return pipeline

        log.info("Overriding compute target if upstream steps use DC...")

        # build and parse the dependency graph
        dependency_graph = _DependencyGraph.from_pipeline(
            pipeline=pipeline, debug=debug
        )

        should_use_dc = dependency_graph.determine_dc_nodes(dc_targets=dc_targets)

        log.info(f"Need to configure the following for DC: {sorted(should_use_dc)}")

        # prepare a mapping from known prod targets to (their corresponding
        # key in the shrike config and) the corresponding DC target
        target_map: Dict[str, Tuple[str, Optional[str]]] = {}
        if self.unified_workspace:
            for prod_target_info in self.config.uw_config.computes:
                if (
                    prod_target_info.DC
                    or prod_target_info.type.lower() == "datafactory"
                ):
                    continue
                prod_target = prod_target_info.name
                dc_target = self._get_uw_computes(
                    prod_target_info.location,
                    prod_target_info.environment,
                    prod_target_info.type,
                    prod_target_info.os,
                    prod_target_info.gpu,
                    True,
                )
                target_map[prod_target] = [prod_target, dc_target]
        else:
            for prod_target in [
                "hdi_prod_target",
                "linux_cpu_prod_target",
                "linux_gpu_prod_target",
                "windows_cpu_prod_target",
                "synapse_prod_target",
            ]:
                if prod_target not in self.config.compute:
                    continue
                target_map[self.config.compute[prod_target]] = [
                    prod_target,
                    self.config.compute.get(
                        prod_target.replace("_prod_", "_dc_"), None
                    ),
                ]

        # move to DC!
        for node_id in should_use_dc:
            node = dependency_graph.nodes[node_id]

            if isinstance(node, Output):

                if node._owner.type == "ScopeComponent":
                    # output datastore has to be cosmos (ADLSg1), while dc_datastore is ADLSg2
                    continue
                if self.unified_workspace:
                    dc_datastore = None
                    prod_datastore = node._datastore
                    for datastore in self.config.uw_config.datastores:
                        if datastore.name == prod_datastore:
                            dc_datastore = self._get_uw_datastores(
                                datastore.location,
                                datastore.environment,
                                True,
                            )
                            break
                if dc_datastore:
                    node.configure(
                        datastore=dc_datastore,
                        output_mode=None,
                    )
                    log.info(f"Configured {node_id} to use datastore {dc_datastore}.")

            elif not hasattr(node.runsettings, "target"):
                # e.g. scope component doesn't need to configure compute target
                log.info(f"Node {node.name} does not have `target`.")
                continue
            elif not isinstance(node.runsettings.target, str):
                log.warn(
                    f"The compute target of node {node.name} is a PipelineParameter. Shrike will not update even if this or the upstream steps use DC. Please update manually."
                )
                continue
            elif node.runsettings.target in dc_targets:
                # nothing to do here
                pass

            elif node.runsettings.target not in target_map:
                log.error(
                    f"Node {node} is using an unknown target: "
                    f"{node.runsettings.target}. "
                    f"Please move this to DC manually, or re-run "
                    f"this after adding the target to the compute config."
                )

            else:
                prod_target_key, dc_target = target_map[node.runsettings.target]
                if dc_target is None:
                    log.error(
                        f"No corresponding DC target is defined for "
                        f"prod target {prod_target_key}. "
                        f"Please move this to DC manually, or re-run "
                        f"this after adding the DC target to the compute config."
                    )
                    continue

                node.runsettings.configure(target=dc_target)
                log.info(
                    f"Updating compute target for {node_id} to {node.runsettings.target}"
                )
        return pipeline

    ################
    ### MAIN/RUN ###
    ################

    def connect(self):
        """Connect to the AML workspace using internal config"""
        # Only call azureml_connect if there is not an already a pre-existing workspace
        try:
            return self.workspace()
        except:
            return azureml_connect(
                aml_subscription_id=self.config.aml.subscription_id,
                aml_resource_group=self.config.aml.resource_group,
                aml_workspace_name=self.config.aml.workspace_name,
                aml_auth=self.config.aml.auth,
                aml_tenant=self.config.aml.tenant,
                aml_force=self.config.aml.force,
            )  # NOTE: this also stores aml workspace in internal global variable

    def _build_pipeline(self) -> Pipeline:
        log.info(f"Building Pipeline [{self.__class__.__name__}]...")
        pipeline_function = self.build(self.config)

        log.info("Creating Pipeline Instance...")
        pipeline = self.pipeline_instance(pipeline_function, self.config)

        if not self.config.run.skip_update_dc and self.is_eyesoff():
            log.info("Overriding compute target if upstream steps use DC...")
            pipeline = self._update_dc_configs(pipeline)

        if not self.config.run.skip_validation:
            log.info("Validating...")
            pipeline.validate(workspace=self.workspace())

        if self.config.run.export:
            log.info(f"Exporting to {self.config.run.export}...")
            with open(self.config.run.export, "w") as export_file:
                export_file.write(pipeline._get_graph_json(workspace=self.workspace()))
        return pipeline

    def _publish_to_endpoint(self, pipeline):
        name = datetime.datetime.now()
        endpoint_name = self.config.run.endpoint_name or self.config.run.experiment_name
        endpoint_description = (
            self.config.run.endpoint_description
            or self.config.run.experiment_description
        )

        pipeline._publish_to_endpoint(
            experiment_name=self.config.run.experiment_name,
            name=name,
            pipeline_endpoint_name=endpoint_name,
            description=endpoint_description,
            workspace=self.workspace(),
        )
        log.info(
            f"Publishing pipeline as endpoint {endpoint_name} with name {name} to workspace {self.workspace()}."
        )

    def _submit_pipeline(self, pipeline: Pipeline) -> Optional[PipelineRun]:
        """Publish and submit the given pipeline.

        Args:
            pipeline: pipeline to be submitted
        """
        if self.config.run.publish:
            self._publish_to_endpoint(pipeline)

        if self.config.run.submit:
            pipeline_tags = self._parse_pipeline_tags()
            pipeline_tags.update({"shrike": __version__})
            pipeline_tags.update(self.repository_info)
            pipeline_tags.update(self._validate_tags(self.extra_tags()))
            log.info(f"Submitting Experiment... [tags={pipeline_tags}]")

            # pipeline_run is of the class "azure.ml.component.run", which
            # is different from "azureml.pipeline.core.PipelineRun"
            pipeline_run = pipeline.submit(
                workspace=self.workspace(),
                experiment_name=self.config.run.experiment_name,
                description=self.config.run.experiment_description,
                display_name=self.config.run.display_name,
                tags=pipeline_tags,
                default_compute_target=None
                if self.unified_workspace
                else self.config.compute.default_compute_target,
                regenerate_outputs=self.config.run.regenerate_outputs,
                continue_on_step_failure=self.config.run.continue_on_failure,
                skip_validation=self.config.run.skip_validation,
            )

            # Forece pipeline_run to be of the class "azureml.pipeline.core.PipelineRun"
            pipeline_run = PipelineRun(
                experiment=pipeline_run._experiment,
                run_id=pipeline_run._id,
            )
            return pipeline_run

        else:
            log.info(
                "Exiting now, if you want to submit please override run.submit=True"
            )
            self.__class__.BUILT_PIPELINE = (
                pipeline  # return so we can have some unit tests done
            )
            return

    @classmethod
    def main(cls):
        """Pipeline helper main function, parses arguments and run pipeline."""
        config_dict = cls._default_config()

        @hydra.main(config_name="default")
        def hydra_run(cfg: DictConfig):
            # merge cli config with default config
            cfg = OmegaConf.merge(config_dict, cfg)

            arg_parser = argparse.ArgumentParser()
            arg_parser.add_argument("--config-dir")
            args, _ = arg_parser.parse_known_args()
            cfg.run.config_dir = os.path.join(
                HydraConfig.get().runtime.cwd, args.config_dir
            )

            log.info("*** CONFIGURATION ***")
            log.info(OmegaConf.to_yaml(cfg))

            # create class instance
            main_instance = cls(cfg)

            # run
            main_instance.run()

        hydra_run()

        return cls.BUILT_PIPELINE  # return so we can have some unit tests done


class _DependencyGraph:
    NodeType = Union[Component, Output]

    graph: Dict[str, List[str]] = {}
    nodes: Dict[str, NodeType] = {}
    debug: bool = False

    def map_to_id(self, item: Union[Pipeline, Component, Output]) -> Optional[str]:
        """Maps the given item to an id."""
        # Reminder: class-wise, pipelines are also components
        extra = f"{item.comment}, " if self.debug and hasattr(item, "comment") else ""
        if isinstance(item, Pipeline):
            return f"pipeline `{item.display_name.replace(' ', '-')}` ({extra}{item._instance_id})"
        if isinstance(item, Component):
            return f"component `{item.display_name.replace(' ', '-')}` ({extra}{item._instance_id})"
        if isinstance(item, Output):
            return f"output `{item.port_name}` of {self.map_to_id(item._owner)}"
        raise ValueError(
            f"Please report this. Could not map {item} to dependency graph id."
        )

    def add_edge(
        self,
        from_node: NodeType,
        to_node: NodeType,
    ) -> None:
        """Add an edge from a dependent node its dependency."""
        for node in [from_node, to_node]:
            if not isinstance(node, (Component, Output)):
                raise ValueError(
                    f"Please report this. "
                    f"Expected a valid node type but was given {node} as node."
                )
            node_id = self.map_to_id(node)
            known_node = self.nodes.get(node_id)
            if known_node is not None and node is not known_node:
                raise ValueError("Report this. Two different nodes with same id.")
            self.nodes[node_id] = node

        ids = [self.map_to_id(node) for node in [from_node, to_node]]
        if ids[0] == ids[1]:
            return
        to_nodes = self.graph.setdefault(ids[0], [])
        if ids[1] not in to_nodes:  # yeah it's a list but short
            to_nodes.append(ids[1])
        self.assert_acyclic()
        log.debug(f"Add edge {ids})")

    def topsorted_ids(self) -> List[str]:
        return toposort_flatten(self.graph)

    def assert_acyclic(self) -> None:
        try:
            self.topsorted_ids()
        except CircularDependencyError as e:
            raise ValueError(
                "Error while constructing dependency graph for DC management."
            ) from e

    @classmethod
    def from_pipeline(cls, pipeline: Pipeline, debug: bool) -> _DependencyGraph:
        graph = _DependencyGraph()
        graph.debug = debug

        def resolve(node: Any) -> Optional[_DependencyGraph.NodeType]:
            if isinstance(node, Output) or isinstance(node, Component):
                return node
            if isinstance(node, Input):
                input_dset = node._dset
                if input_dset is None:
                    return None
                return resolve(input_dset)
            if isinstance(node, PipelineParameter):
                default_value = node.default_value
                if default_value is None or default_value is node:
                    # can't resolve this
                    return None
                return resolve(default_value)
            return None

        for subpipeline in pipeline._expand_pipeline_to_pipelines():
            # For DC, we only need to adapt `Output` and `Component`
            # components. The `Output`s of pipelines do not coincide
            # with the `Output`s of components, hence we have to
            # insert those into our dependency graph:
            for output_name, output in subpipeline.outputs.items():
                linked_output = subpipeline._outputs_mapping[output_name]
                graph.add_edge(output, linked_output)
            # However, it seems we can skip the pipeline inputs.

        for component in pipeline._expand_pipeline_nodes():
            for output in component.outputs.values():
                graph.add_edge(output, component)
            for input_ in component.inputs.values():
                corresponding_node = resolve(input_)
                if corresponding_node is not None:
                    graph.add_edge(component, corresponding_node)
        return graph

    def determine_dc_nodes(self, dc_targets: List[str]) -> Set[str]:
        """Determine which downstream components/outputs should be configured for DC.

        Args:
            dc_targets: known DC targets (used to determine which nodes
               are already configured for DC)
        Returns:
            list of node ids for all nodes that should be configured for
            dc (or have already been configured for dc)
        """
        should_use_dc: Set[str] = set()
        for node_id in self.topsorted_ids():

            # use dc, if upstream used dc
            for dependency in self.graph.get(node_id, []):
                if dependency in should_use_dc:
                    should_use_dc.add(node_id)
                    continue

            item = self.nodes[node_id]

            if not isinstance(item, Component):
                # If users choose to move an output to iso datastore out of
                # nowhere, we ignore this. We're only caring about cases where
                # the (`apply_*` method caused the) run settings of components
                # to be configured for DC - and fix downstream components.
                continue

            if not hasattr(item.runsettings, "target") or not isinstance(
                item.runsettings.target, str
            ):
                # e.g. Scope components do not configure compute targets
                # e.g. if `target` is a PipelineParameter, it returns None
                # and no need to update
                continue

            if item.runsettings.target in dc_targets:
                should_use_dc.add(node_id)

        return should_use_dc



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
"""
Avoid breaking changes to `from shrike.pipeline.module_helper import AMLModuleLoader`
"""

from .v1 import AMLModuleLoader
from .module_helper_base import *



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Pipeline helper class to create pipelines loading modules from a flexible manifest.
"""
from shrike.pipeline.module_helper_base import AMLModuleLoaderBase

try:
    from azure.ai.ml import load_component, MLClient
except ImportError as error:
    raise ImportError(
        f"{error.msg}. Please install v2 dependency `pip install azure-ai-ml`."
    )
import os
import logging
from packaging.specifiers import SpecifierSet
from packaging.version import Version, InvalidVersion

from shrike.pipeline.v2.aml_connect import current_workspace, get_registry_client

log = logging.getLogger(__name__)


class AMLModuleLoaderV2(AMLModuleLoaderBase):
    """Helper class to load modules from within an AMLPipelineHelper in dpv2."""

    def load_local_module_helper(self, module_spec_path):
        return load_component(source=module_spec_path)

    def solve_module_version_and_load(
        self, module_name, module_version, module_cache_key, registry=None
    ):
        """Loads module class if exists
        Args:
            module_name (str): name of the module to load
            module_version (str): version of the module to load
            module_cache_key (str): cache key of the module after loading
            registy (str): registry name if loading from a registry
        """
        client = (
            current_workspace()
            if registry is None
            else get_registry_client(registry_name=registry)
        )
        module_version = self.solve_module_version(
            module_name, module_version, client=client
        )

        # TODO(1989862): change when component support get default version
        if module_version is not None:
            loaded_module_class = client.components.get(
                name=module_name,
                version=module_version,
            )
        else:
            log.info(
                f"No version specified for {module_name}, trying to load with 'latest' label."
            )
            loaded_module_class = client.components.get(
                name=module_name,
                label="latest",
            )

        self.put_in_cache(module_cache_key, loaded_module_class)
        return loaded_module_class

    def solve_module_version(self, module_name, module_version, client: MLClient):
        if module_version is None:
            return module_version

        try:
            module_version_PEP440 = Version(module_version)
            if str(module_version_PEP440) != module_version:
                log.warning(
                    "We suggest adopting PEP440 versioning for your component {module_name}!"
                )

        except InvalidVersion as e:
            log.info(f"{module_version} is a version constraint. Try to solve it...")

            spec = SpecifierSet(module_version)
            components = client.components.list(name=module_name)

            versions = []
            for component in components:
                version = component.version
                version_PEP440 = Version(version)
                if str(version_PEP440) != version:
                    log.warning(
                        f"Version {version} does not follow PEP440 versioning, skipping ..."
                    )
                else:
                    versions.append(version_PEP440)

            compatible_versions = list(spec.filter(versions))
            if compatible_versions:
                module_version = max(compatible_versions)
                log.info(f"Solved version for {module_name} is {module_version}.")

            else:
                raise ValueError(
                    f"No version exists for the constraint {module_version}. Existing versions: {versions}"
                )

        return str(module_version)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.


"""
Setup script for this Python package.
https://docs.python.org/3/distutils/setupscript.html
"""


import pathlib
from setuptools import setup, find_packages
from shrike import __version__
import glob


def versions_in_requirements(file):
    lines = file.read().splitlines()
    versions = [
        line
        for line in lines
        # https://stackoverflow.com/a/2405300
        if not line.isspace() and "--" not in line
    ]
    return list(versions)


HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

with open(HERE / "requirements/requirements-logging.txt") as f:
    required_logging = versions_in_requirements(f)

with open(HERE / "requirements/requirements-pipeline.txt") as f:
    required_pipeline = versions_in_requirements(f)

with open(HERE / "requirements/requirements-distributed.txt") as f:
    required_distributed = versions_in_requirements(f)

with open(HERE / "requirements/requirements-build.txt") as f:
    required_build = versions_in_requirements(f)

with open(HERE / "requirements/requirements-dev.txt") as f:
    required_dev = versions_in_requirements(f)


def component_files():
    return glob.glob("shrike/components/**/**")


setup(
    name="shrike",
    version=__version__,
    description="Python utilities for compliant Azure machine learning",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/ai-platform-ml-platform/shrike",
    author="AML Data Science",
    author_email="aml-ds@microsoft.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    packages=find_packages(include=["shrike*"]),
    include_package_data=True,
    install_requires=required_logging,
    extras_require={
        "pipeline": required_pipeline,
        "distributed": required_distributed,
        "build": required_build,
        "dev": required_dev,
    },
    # https://stackoverflow.com/a/48777286
    python_requires=">=3.6",
    package_data={"components": component_files()},
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from azureml.core import Experiment, ScriptRunConfig, Environment
from azureml.core.environment import CondaDependencies
import os
from omegaconf import OmegaConf
import hydra
from hydra.core.hydra_config import HydraConfig
from shrike.pipeline.aml_connect import azureml_connect
from shrike._core import (
    is_eyesoff_helper,
    b64_decode,
    b64_encode,
    experimental,
    O365_FEED,
)
from shrike.pipeline import AMLPipelineHelper


def merge_serialize_config_with_default(serialized_config):
    default_config = AMLPipelineHelper._default_config()
    cfg = OmegaConf.merge(
        default_config, OmegaConf.structured(b64_decode(serialized_config))
    )
    return cfg


def get_or_create_env(ws, env_name, is_eyesoff):
    try:
        env = Environment.get(ws, env_name)
    except Exception:
        env = Environment(env_name)
        conda_dep = CondaDependencies.create(
            pip_packages=[
                "ray>=1.7.0",
                "shrike[pipeline]>=1.28.0",
            ],
            python_version="3.8",
            pip_indexurl=O365_FEED if is_eyesoff else None,
        )
        env.python.conda_dependencies = conda_dep
    return env


@experimental()
@hydra.main(config_name="default")
def ray_actor_on_shrike(cfg):
    default_config = AMLPipelineHelper._default_config()
    cfg = OmegaConf.merge(default_config, cfg)
    env_name = "ray_env"

    aml = cfg.aml
    ws = azureml_connect(
        aml_subscription_id=aml.subscription_id,
        aml_resource_group=aml.resource_group,
        aml_workspace_name=aml.workspace_name,
        aml_auth=aml.get("auth", "interactive"),
        aml_tenant=aml.get("tenant", None),
        aml_force=aml.get("force", False),
    )
    env = get_or_create_env(
        ws, env_name, is_eyesoff_helper(aml.get("tenant", None), aml.subscription_id)
    )
    compute_target = cfg.compute.default_compute_target

    script = os.path.join(HydraConfig.get().runtime.cwd, cfg.run.script)
    with open(script) as file:
        contents = file.read()

    script_config = ScriptRunConfig(
        source_directory=".",
        command=["echo", "'" + contents + "'"]
        + "> script.py && python script.py --serialized_config".split()
        + [b64_encode(OmegaConf.to_yaml(cfg))],
        compute_target=compute_target,
        environment=env,
    )
    exp = Experiment(ws, cfg.run.experiment_name)
    run = exp.submit(script_config)
    print(run.get_portal_url())
    run.wait_for_completion()
